#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import re
from datetime import datetime

from lxml.builder import ElementMaker
from django.utils.six import string_types

from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap, ns_xsi
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.services.subset import Trim, Slice, is_temporal, all_axes
from eoxserver.services.gml.v32.encoders import (
    ns_gml, ns_gmlcov, ns_om, ns_eop, GML, GMLCOV, OM, EOP
)
from eoxserver.services.ows.common.v20.encoders import ns_xlink, ns_ows, OWS
from eoxserver.services.exceptions import (
    InvalidSubsettingException, InvalidAxisLabelException,
    NoSuchFieldException, InvalidFieldSequenceException,
    InterpolationMethodNotSupportedException, InvalidScaleFactorException,
    InvalidScaleExtentException, ScaleAxisUndefinedException
)


# namespace declarations
ns_ogc = NameSpace("http://www.opengis.net/ogc", "ogc")
ns_wcs = NameSpace("http://www.opengis.net/wcs/2.0", "wcs")
ns_crs = NameSpace("http://www.opengis.net/wcs/crs/1.0", "crs")
ns_rsub = NameSpace("http://www.opengis.net/wcs/range-subsetting/1.0", "rsub")
ns_eowcs = NameSpace("http://www.opengis.net/wcs/wcseo/1.0", "wcseo",
                     "http://schemas.opengis.net/wcs/wcseo/1.0/wcsEOAll.xsd")
ns_swe = NameSpace("http://www.opengis.net/swe/2.0", "swe")
ns_int = NameSpace("http://www.opengis.net/wcs/interpolation/1.0", "int")
ns_scal = NameSpace("http://www.opengis.net/wcs/scaling/1.0", "scal")

# namespace map
nsmap = NameSpaceMap(
    ns_xlink, ns_ogc, ns_ows, ns_gml, ns_gmlcov, ns_wcs, ns_crs, ns_rsub,
    ns_eowcs, ns_om, ns_eop, ns_swe, ns_int, ns_scal
)

# Element factories

WCS = ElementMaker(namespace=ns_wcs.uri, nsmap=nsmap)
CRS = ElementMaker(namespace=ns_crs.uri, nsmap=nsmap)
EOWCS = ElementMaker(namespace=ns_eowcs.uri, nsmap=nsmap)
SWE = ElementMaker(namespace=ns_swe.uri, nsmap=nsmap)
INT = ElementMaker(namespace=ns_int.uri, nsmap=nsmap)


SUBSET_RE = re.compile(r'(\w+)\(([^,]*)(,([^)]*))?\)')
SCALEAXIS_RE = re.compile(r'(\w+)\(([^)]*)\)')
SCALESIZE_RE = SCALEAXIS_RE
SCALEEXTENT_RE = re.compile(r'(\w+)\(([^:]*):([^)]*)\)')


class RangeSubset(list):
    def get_band_indices(self, range_type, offset=0):
        current_idx = -1
        all_bands = range_type[:]

        for subset in self:
            if isinstance(subset, string_types):
                # slice, i.e single band
                start = stop = subset

            else:
                start, stop = subset

            start_idx = self._find(all_bands, start)
            if start != stop:
                stop_idx = self._find(all_bands, stop)
                if stop_idx <= start_idx:
                    raise IllegalFieldSequenceException(
                        "Invalid interval '%s:%s'." % (start, stop), start
                    )

                # expand interval to indices
                for i in range(start_idx, stop_idx+1):
                    yield i + offset

            else:
                # return the item
                yield start_idx + offset


    def _find(self, all_bands, name):
        for i, band in enumerate(all_bands):
            if band.identifier == name:
                return i
        raise NoSuchFieldException("Field '%s' does not exist." % name, name)


class Scale(object):
    """ Abstract base class for all Scaling operations.
    """
    def __init__(self, axis):
        self.axis = axis


class ScaleAxis(Scale):
    """ Scale a single axis by a specific value.
    """
    def __init__(self, axis, scale):
        super(ScaleAxis, self).__init__(axis)
        self.scale = scale


class ScaleSize(Scale):
    """ Scale a single axis to a specific size.
    """
    def __init__(self, axis, size):
        super(ScaleSize, self).__init__(axis)
        self.size = size


class ScaleExtent(Scale):
    """ Scale a single axis to a specific extent.
    """
    def __init__(self, axis, low, high):
        super(ScaleExtent, self).__init__(axis)
        self.low = low
        self.high = high


class SectionsMixIn(object):
    """ Mix-in for request decoders that use sections.
    """

    def section_included(self, *sections):
        """ See if one of the sections is requested.
        """
        if not self.sections:
            return True

        requested_sections = [s.lower() for s in self.sections]

        for section in map(lambda s: s.lower(), sections):
            section = section.lower()
            if "all" in requested_sections or section in requested_sections:
                return True

        return False


def parse_subset_kvp(string):
    """ Parse one subset from the WCS 2.0 KVP notation.
    """

    try:
        match = SUBSET_RE.match(string)
        if not match:
            raise Exception("Could not parse input subset string.")

        axis = match.group(1)
        parser = get_parser_for_axis(axis)

        if match.group(4) is not None:
            return Trim(
                axis, parser(match.group(2)), parser(match.group(4))
            )
        else:
            return Slice(axis, parser(match.group(2)))
    except InvalidAxisLabelException:
        raise
    except Exception as e:
        raise InvalidSubsettingException(str(e))


def parse_range_subset_kvp(string):
    """ Parse a rangesubset structure from the WCS 2.0 KVP notation.
    """

    rangesubset = RangeSubset()
    for item in string.split(","):
        if ":" in item:
            rangesubset.append(item.split(":"))
        else:
            rangesubset.append(item)

    return rangesubset


def parse_scaleaxis_kvp(string):
    """ Parses the KVP notation of a single scale axis.
    """

    match = SCALEAXIS_RE.match(string)
    if not match:
        raise Exception("Could not parse input scale axis string.")

    axis = match.group(1)
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        value = float(match.group(2))
    except ValueError:
        raise InvalidScaleFactorException(match.group(2))

    return ScaleAxis(axis, value)


def parse_scalesize_kvp(string):
    """ Parses the KVP notation of a single scale size.
    """

    match = SCALESIZE_RE.match(string)
    if not match:
        raise Exception("Could not parse input scale size string.")

    axis = match.group(1)
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        value = int(match.group(2))
    except ValueError:
        raise InvalidScaleFactorException(match.group(2))

    return ScaleSize(axis, value)


def parse_scaleextent_kvp(string):
    """ Parses the KVP notation of a single scale extent.
    """

    match = SCALEEXTENT_RE.match(string)
    if not match:
        raise Exception("Could not parse input scale extent string.")

    axis = match.group(1)
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        low = int(match.group(2))
        high = int(match.group(3))
    except ValueError:
        raise InvalidScaleFactorException(match.group(3))

    if low >= high:
        raise InvalidScaleExtentException(low, high)

    return ScaleExtent(axis, low, high)


def parse_subset_xml(elem):
    """ Parse one subset from the WCS 2.0 XML notation. Expects an lxml.etree
        Element as parameter.
    """

    try:
        dimension = elem.findtext(ns_wcs("Dimension"))
        parser = get_parser_for_axis(dimension)
        if elem.tag == ns_wcs("DimensionTrim"):
            return Trim(
                dimension,
                parser(elem.findtext(ns_wcs("TrimLow"))),
                parser(elem.findtext(ns_wcs("TrimHigh")))
            )
        elif elem.tag == ns_wcs("DimensionSlice"):
            return Slice(
                dimension,
                parser(elem.findtext(ns_wcs("SlicePoint")))
            )
    except Exception as e:
        raise InvalidSubsettingException(str(e))


SUPPORTED_INTERPOLATIONS = (
    "average", "nearest-neighbour", "bilinear", "cubic", "cubic-spline",
    "lanczos", "mode"
)

def parse_interpolation(raw):
    """ Returns a unified string denoting the interpolation method used.
    """
    if raw.startswith("http://www.opengis.net/def/interpolation/OGC/1/"):
        raw = raw[len("http://www.opengis.net/def/interpolation/OGC/1/"):]
        value = raw.lower()
    else:
        value = raw.lower()

    if value not in SUPPORTED_INTERPOLATIONS:
        raise InterpolationMethodNotSupportedException(
            "Interpolation method '%s' is not supported." % raw
        )
    return value


def parse_range_subset_xml(elem):
    """ Parse a rangesubset structure from the WCS 2.0 XML notation.
    """

    rangesubset = RangeSubset()

    for child in elem:
        item = child[0]
        if item.tag == ns_rsub("RangeComponent"):
            rangesubset.append(item.text)
        elif item.tag == ns_rsub("RangeInterval"):
            rangesubset.append((
                item.findtext(ns_rsub("startComponent")),
                item.findtext(ns_rsub("endComponent"))
            ))

    return rangesubset


def parse_scaleaxis_xml(elem):
    """ Parses the XML notation of a single scale axis.
    """

    axis = elem.findtext(ns_scal("axis"))
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        raw = elem.findtext(ns_scal("scaleFactor"))
        value = float(raw)
    except ValueError:
        InvalidScaleFactorException(raw)

    return ScaleAxis(axis, value)


def parse_scalesize_xml(elem):
    """ Parses the XML notation of a single scale size.
    """

    axis = elem.findtext(ns_scal("axis"))
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        raw = elem.findtext(ns_scal("targetSize"))
        value = int(raw)
    except ValueError:
        InvalidScaleFactorException(raw)

    return ScaleSize(axis, value)


def parse_scaleextent_xml(elem):
    """ Parses the XML notation of a single scale extent.
    """

    axis = elem.findtext(ns_scal("axis"))
    if axis not in all_axes:
        raise ScaleAxisUndefinedException(axis)
    try:
        raw_low = elem.findtext(ns_scal("low"))
        raw_high = elem.findtext(ns_scal("high"))
        low = int(raw_low)
        high = int(raw_high)
    except ValueError:
        InvalidScaleFactorException(raw_high)

    if low >= high:
        raise InvalidScaleExtentException(low, high)

    return ScaleExtent(axis, low, high)


def float_or_star(value):
    """ Parses a string value that is either a floating point value or the '*'
        character. Raises a `ValueError` if no float could be parsed.
    """

    if value == "*":
        return None
    try:
        return float(value)
    except ValueError:
        raise ValueError("could not convert string to float: '%s'" % value)


def parse_quoted_temporal(value):
    """ Parses a quoted temporal value.
    """

    if value == "*":
        return None

    if not value[0] == '"' and not value[-1] == '"':
        raise ValueError(
            "Temporal value needs to be quoted with double quotes."
        )

    return parse_iso8601(value[1:-1])


def get_parser_for_axis(axis):
    """ Returns the correct parsing function for the given axis.
    """

    if is_temporal(axis):
        return parse_quoted_temporal
    else:
        return float_or_star



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import re
import logging
import hashlib
import json

from django.core.cache import caches, InvalidCacheBackendError
from django.utils import timezone
from keystoneclient import exceptions as ksexceptions
from keystoneclient.v2_0 import client as ksclient_v2
from keystoneclient.v3 import client as ksclient_v3
from swiftclient.exceptions import ClientException
from django.utils.six.moves.urllib import parse

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.backends.storage_auths import BaseStorageAuthHandler

logger = logging.getLogger(__name__)


VERSIONFUL_AUTH_PATH = re.compile(r'v[2-3](?:\.0)?$')
AUTH_VERSIONS_V2 = ('2.0', '2', 2)
AUTH_VERSIONS_V3 = ('3.0', '3', 3)


def get_keystone_client(auth_url, user, key, os_options, **kwargs):
    """
    Authenticate against a keystone server.
    We are using the keystoneclient library for authentication.
    """

    insecure = kwargs.get('insecure', False)
    timeout = kwargs.get('timeout', None)
    auth_version = kwargs.get('auth_version', None)
    debug = logger.isEnabledFor(logging.DEBUG)

    # Add the version suffix in case of versionless Keystone endpoints. If
    # auth_version is also unset it is likely that it is v3
    if not VERSIONFUL_AUTH_PATH.match(
            parse.urlparse(auth_url).path.rstrip('/').rsplit('/', 1)[-1]):
        # Normalize auth_url to end in a slash because urljoin
        auth_url = auth_url.rstrip('/') + '/'
        if auth_version and auth_version in AUTH_VERSIONS_V2:
            auth_url = parse.urljoin(auth_url, "v2.0")
        else:
            auth_url = parse.urljoin(auth_url, "v3")
            auth_version = '3'
        logger.debug("Versionless auth_url - using %s as endpoint" % auth_url)

    # Legacy default if not set
    if auth_version is None:
        auth_version = '2'

    ksclient = None
    if auth_version in AUTH_VERSIONS_V3:
        if ksclient_v3 is not None:
            ksclient = ksclient_v3
    else:
        if ksclient_v2 is not None:
            ksclient = ksclient_v2

    if ksclient is None:
        raise ClientException('''
Auth versions 2.0 and 3 require python-keystoneclient, install it or use Auth
version 1.0 which requires ST_AUTH, ST_USER, and ST_KEY environment
variables to be set or overridden with -A, -U, or -K.''')

    try:
        client = ksclient.Client(
            username=user,
            password=key,
            token=os_options.get('auth_token'),
            tenant_name=os_options.get('tenant_name'),
            tenant_id=os_options.get('tenant_id'),
            user_id=os_options.get('user_id'),
            user_domain_name=os_options.get('user_domain_name'),
            user_domain_id=os_options.get('user_domain_id'),
            project_name=os_options.get('project_name'),
            project_id=os_options.get('project_id'),
            project_domain_name=os_options.get('project_domain_name'),
            project_domain_id=os_options.get('project_domain_id'),
            debug=debug,
            cacert=kwargs.get('cacert'),
            cert=kwargs.get('cert'),
            key=kwargs.get('cert_key'),
            auth_url=auth_url, insecure=insecure, timeout=timeout)
    except ksexceptions.Unauthorized:
        msg = 'Unauthorized. Check username, password and tenant name/id.'
        if auth_version in AUTH_VERSIONS_V3:
            msg = ('Unauthorized. Check username/id, password, '
                   'tenant name/id and user/tenant domain name/id.')
        raise ClientException(msg)
    except ksexceptions.AuthorizationFailure as err:
        raise ClientException('Authorization Failure. %s' % err)

    return client


def get_endpoint_url_and_token(client, os_options):
    service_type = os_options.get('service_type') or 'object-store'
    endpoint_type = os_options.get('endpoint_type') or 'publicURL'
    try:
        filter_kwargs = {}
        if os_options.get('region_name'):
            filter_kwargs['attr'] = 'region'
            filter_kwargs['filter_value'] = os_options['region_name']
        endpoint = client.service_catalog.url_for(
            service_type=service_type,
            endpoint_type=endpoint_type,
            **filter_kwargs)
    except ksexceptions.EndpointNotFound:
        raise ClientException('Endpoint for %s not found - '
                              'have you specified a region?' % service_type)
    return endpoint, client.auth_token


def get_auth_expires_at(client):
    return parse_iso8601(client.auth_ref['expires_at'])


class KeystoneStorageAuthHandler(BaseStorageAuthHandler):
    name = 'keystone'

    def _get_url_and_token(self):
        url, token = self.get_auth()
        return url, token

    def get_auth(self):
        os_options = {
            key: self.parameters.get(key)
            for key in [
                'tenant_id', 'auth_token', 'service_type', 'endpoint_type',
                'tenant_name', 'object_storage_url', 'region_name',
                'service_username', 'service_project_name', 'service_key'
            ]
        }

        try:
            cache = caches['default']
        except InvalidCacheBackendError:
            cache = None

        if cache:
            base_key = '%s' % hashlib.sha256(json.dumps(dict(
                url=self.url,
                user=self.parameters.get('username'),
                key=self.parameters.get('password'),
                os_options=os_options,
            )).encode('utf-8')).hexdigest()

            token_key = 'keystone_auth_token_%s' % base_key
            url_key = 'keystone_storage_url_%s' % base_key
            token = cache.get(token_key)
            url = cache.get(url_key)
            if url and token:
                logger.debug(
                    'Using cached swift storage URL and access token'
                )
                return url, token

        logger.debug(
            'Fetching swift storage URL and access token'
        )
        client = get_keystone_client(
            self.url,
            user=self.parameters.get('username'),
            key=self.parameters.get('password'),
            os_options=os_options,
        )
        expires_in = (
            get_auth_expires_at(client) - timezone.now()
        ).total_seconds()
        url, token = get_endpoint_url_and_token(client, os_options)

        if cache:
            logger.debug(
                'Caching swift storage URL and access token. '
                'Valid for %f seconds.'
                % expires_in
            )
            cache.set(url_key, url, expires_in)
            cache.set(token_key, token, expires_in)

        return url, token

    def get_vsi_env(self):
        url, token = self._get_url_and_token()
        return {
            'SWIFT_STORAGE_URL': url,
            'SWIFT_AUTH_TOKEN': token,
        }



#-------------------------------------------------------------------------------
#
#  WPS Literal Data - allowed values - debuging unit-tests
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=missing-docstring

from unittest import TestCase, main
from datetime import time, date, datetime, timedelta
from eoxserver.services.ows.wps.parameters import (
    AllowedAny, AllowedEnum, AllowedRange, AllowedRangeCollection,
    Integer, Double, String, Duration, Date, Time, DateTime,
)

from django.utils.six import text_type
#------------------------------------------------------------------------------

class BaseTestMixin(object):
    # pylint: disable=too-few-public-methods
    def test(self):
        for val in self.accepted:
            try:
                self.assertTrue(self.domain.check(val))
                self.assertTrue(val is self.domain.verify(val))
            except:
                print ("\n%s: value: %r" % (type(self).__name__, val))
                raise
        for val in self.rejected:
            try:
                self.assertFalse(self.domain.check(val))
                self.assertRaises(ValueError, self.domain.verify, val)
            except:
                print ("\n%s: value: %r" % (type(self).__name__, val))
                raise

#------------------------------------------------------------------------------

class TestAllowedAny(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedAny()
        self.accepted = [
            1.0,
            10,
            datetime.now(),
            timedelta(days=10, seconds=100),
            'test',
        ]
        self.rejected = []

#------------------------------------------------------------------------------

class TestAllowedEnumFloat(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedEnum([0.0, 0.5, 1.0])
        self.accepted = [1.0, 0]
        self.rejected = [-1]


class TestAllowedEnumFloat2(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedEnum([0.0, 0.5, 1.0], dtype=Double)
        self.accepted = [1.0, 0]
        self.rejected = [-1]


class TestAllowedEnumFloat3(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedEnum([0.0, 0.5, 1.0], dtype=float)
        self.accepted = [1.0, 0]
        self.rejected = [-1]


class TestAllowedEnumInt(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedEnum([0, 2, 3], dtype=Integer)
        self.accepted = [2, 0]
        self.rejected = [1]


class TestAllowedEnumInt2(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedEnum([0, 2, 3], dtype=int)
        self.accepted = [2, 0]
        self.rejected = [1]


class TestAllowedEnumString(TestCase, BaseTestMixin):
    def setUp(self):
        enum = ['John', 'James', 'Jeffrey', 'Jacob', 'Jerry']
        self.domain = AllowedEnum(enum, dtype=String)
        self.accepted = ['John', 'Jacob', 'Jerry']
        self.rejected = ['Alex', '']


class TestAllowedEnumString2(TestCase, BaseTestMixin):
    def setUp(self):
        enum = ['John', 'James', 'Jeffrey', 'Jacob', 'Jerry']
        self.domain = AllowedEnum(enum, dtype=str)
        self.accepted = ['John', 'Jacob', 'Jerry']
        self.rejected = ['Alex', '']


class TestAllowedEnumString3(TestCase, BaseTestMixin):
    def setUp(self):
        enum = ['John', 'James', 'Jeffrey', 'Jacob', 'Jerry']
        self.domain = AllowedEnum(enum, dtype=text_type)
        self.accepted = ['John', 'Jacob', 'Jerry']
        self.rejected = ['Alex', '']


class TestAllowedEnumDate(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['2014-01-01', '2014-02-01', '2014-03-01']
        self.domain = AllowedEnum(vlist, dtype=Date)
        self.accepted = [vlist[1], Date.parse(vlist[0])]
        self.rejected = [Date.parse('2014-01-02')]


class TestAllowedEnumDate2(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['2014-01-01', '2014-02-01', '2014-03-01']
        self.domain = AllowedEnum(vlist, dtype=date)
        self.accepted = [vlist[1], Date.parse(vlist[0])]
        self.rejected = [Date.parse('2014-01-02')]


class TestAllowedEnumTime(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['05:30', '08:20', '16:18']
        self.domain = AllowedEnum(vlist, dtype=Time)
        self.accepted = [vlist[1], Time.parse(vlist[0])]
        self.rejected = [Time.parse('19:20')]


class TestAllowedEnumTime2(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['05:30', '08:20', '16:18']
        self.domain = AllowedEnum(vlist, dtype=time)
        self.accepted = [vlist[1], Time.parse(vlist[0])]
        self.rejected = [Time.parse('19:20')]


class TestAllowedEnumDateTime(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = [
            '2014-01-01T09:30:21Z',
            '2014-02-01T18:20:15Z',
            '2014-03-01T12:15:02Z',
        ]
        self.domain = AllowedEnum(vlist, dtype=DateTime)
        self.accepted = [vlist[1], DateTime.parse(vlist[0])]
        self.rejected = [DateTime.parse('2014-01-01T12:30:00Z')]


class TestAllowedEnumDateTime2(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = [
            '2014-01-01T09:30:21Z',
            '2014-02-01T18:20:15Z',
            '2014-03-01T12:15:02Z',
        ]
        self.domain = AllowedEnum(vlist, dtype=datetime)
        self.accepted = [vlist[1], DateTime.parse(vlist[0])]
        self.rejected = [DateTime.parse('2014-01-01T12:30:00Z')]


class TestAllowedEnumDuration(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['P1Y', 'P26DT1M', 'PT25M16S']
        self.domain = AllowedEnum(vlist, dtype=Duration)
        self.accepted = [vlist[1], Duration.parse(vlist[0])]
        self.rejected = [Duration.parse('P7DT15H8M')]


class TestAllowedEnumDuration2(TestCase, BaseTestMixin):
    def setUp(self):
        vlist = ['P1Y', 'P26DT1M', 'PT25M16S']
        self.domain = AllowedEnum(vlist, dtype=timedelta)
        self.accepted = [vlist[1], Duration.parse(vlist[0])]
        self.rejected = [Duration.parse('P7DT15H8M')]

#------------------------------------------------------------------------------

class TestAllowedRangeFloat(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0)
        self.accepted = [0.5, 0.0, 1.0]
        self.rejected = [-1.0, 2.0]


class TestAllowedRangeFloat2(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, dtype=Double)
        self.accepted = [0.5, 0.0, 1.0]
        self.rejected = [-1.0, 2.0]


class TestAllowedRangeFloat3(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, dtype=float)
        self.accepted = [0.5, 0.0, 1.0]
        self.rejected = [-1.0, 2.0]


class TestAllowedRangeUnboundMin(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(None, 1.0)
        self.accepted = ['-inf', -1.0, 0.5, 0.0, 1.0]
        self.rejected = [2.0]


class TestAllowedRangeUnboundMax(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, None)
        self.accepted = ['+inf', 0.5, 0.0, 1.0, 2.0]
        self.rejected = [-1.0]


class TestAllowedRangeFloatClosed(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, 'closed')
        self.accepted = [0.5, 0.0, 1.0]
        self.rejected = [-1.0, 2.0]


class TestAllowedRangeFloatOpen(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, 'open')
        self.accepted = [0.5]
        self.rejected = [0.0, 1.0, -1.0, 2.0]


class TestAllowedRangeFloatOpenClosed(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, 'open-closed')
        self.accepted = [0.5, 1.0]
        self.rejected = [0.0, -1.0, 2.0]


class TestAllowedRangeFloatClosedOpen(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, 'closed-open')
        self.accepted = [0.5, 0.0]
        self.rejected = [1.0, -1.0, 2.0]


class TestAllowedRangeInt(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0, 10, dtype=Integer)
        self.accepted = [0, 5, 10]
        self.rejected = [-1, 12]


class TestAllowedRangeIntClosed(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0, 10, 'open', dtype=Integer)
        self.accepted = [5]
        self.rejected = [0, 10, -1, 12]


class TestAllowedRangeDateClosed(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            '2014-02-01', '2014-03-01', 'closed', dtype=Date
        )
        self.accepted = [
            '2014-02-15',
            '2014-02-01',
            Date.parse('2014-03-01'),
        ]
        self.rejected = [
            '2014-01-02',
            Date.parse('2014-01-02'),
        ]


class TestAllowedRangeDateOpen(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            '2014-02-01', '2014-03-01', 'open', dtype=Date
        )
        self.accepted = [
            '2014-02-15',
        ]
        self.rejected = [
            '2014-02-01',
            Date.parse('2014-03-01'),
            Date.parse('2014-01-02'),
        ]


class TestAllowedRangeDateClosedOpen(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange('10:00', '15:30', 'closed-open', dtype=Time)
        self.accepted = ['10:00', '12:15']
        self.rejected = [Time.parse('09:00'), '15:30', '18:12']


class TestAllowedRangeDateOpenClosed(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange('10:00', '15:30', 'open-closed', dtype=Time)
        self.accepted = ['12:15', '15:30']
        self.rejected = [Time.parse('09:00'), '10:00', '18:12']


class TestAllowedRangeDateTime(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            '2014-02-01T09:30:21Z', '2014-03-01T18:20:15Z', dtype=DateTime
        )
        self.accepted = [
            '2014-02-01T09:30:21Z',
            '2014-02-15T00:00:00Z',
            '2014-03-01T18:20:15Z',
        ]
        self.rejected = [
            '2014-01-01T00:00:00Z',
            DateTime.parse('2014-04-01T00:00:00Z'),
        ]


class TestAllowedRangeDuration(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange('-P1DT1H', 'P1M0DT30M', dtype=Duration)
        self.accepted = [
            Duration.parse('-P1DT1H'),
            'P0D',
            'P1M0DT30M',
        ]
        self.rejected = [
            Duration.parse('-P2DT18H'),
            'P1Y',
        ]

#------------------------------------------------------------------------------

class TestAllowedRangeDiscrFloat(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0.0, 1.0, spacing='0.1')
        self.accepted = [0.5, 0.0, 1.0]
        self.rejected = [0.55, -1.0, 2.0]


class TestAllowedRangeDiscrInt(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(0, 10, spacing='2', dtype=int)
        self.accepted = [4, 0, 10]
        self.rejected = [5, -1, 12]


class TestAllowedRangeDiscrDuration(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            timedelta(0, 3600, 0), 'PT5H', 'open-closed', spacing='PT1H',
            dtype=timedelta
        )
        self.accepted = [
            Duration.parse('PT2H'),
            timedelta(0, 3*3600, 0),
            'PT4H',
            'PT5H',
            'PT2H0.000001S',
        ]
        self.rejected = [
            'PT30M',
            'PT1H',
            'PT2H30M',
        ]


class TestAllowedRangeDiscrTime(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            time(9, 30, 0, 0), '16:30', 'closed', spacing='PT1H', dtype=time
        )
        self.accepted = ['09:30', '10:30', '15:30', '16:30']
        self.rejected = ['09:00', '10:00', '16:25', '16:35']


class TestAllowedRangeDiscrDate(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            date(2014, 1, 1), '2014-01-21', 'closed', spacing='P5D', dtype=date
        )
        self.accepted = ['2014-01-01', '2014-01-06', '2014-01-16', '2014-01-21']
        self.rejected = ['2013-12-31', '2014-01-10', '2014-01-31']


class TestAllowedRangeDiscrDateTime(TestCase, BaseTestMixin):
    def setUp(self):
        self.domain = AllowedRange(
            datetime(2014, 1, 1, 10, 30, 0, 0, DateTime.UTC),
            '2014-01-10T10:30Z', 'closed', spacing='P1D', dtype=datetime
        )
        self.accepted = [
            '2014-01-01T10:30Z',
            '2014-01-02T10:30Z',
            '2014-01-05T10:30Z',
            '2014-01-07T10:30Z',
            '2014-01-09T10:30Z',
            '2014-01-10T10:30Z',
        ]
        self.rejected = [
            '2013-12-31T10:30Z',
            '2014-01-01T11:00Z',
            '2014-01-05T00:00Z',
            '2014-01-11T10:30Z',
        ]

#------------------------------------------------------------------------------


class TestAllowedRangeCollectionFloat(TestCase, BaseTestMixin):
    def setUp(self):
        try:
            # Python 2
            xrange
        except NameError:
            # Python 3, xrange is now named range
            xrange = range
        self.domain = AllowedRangeCollection(
            AllowedRange(None, -5.0, 'open'),
            AllowedRange(6.0, None, spacing=3.0),
            AllowedEnum(xrange(-4, 0)),
            AllowedEnum(range(0, 6, 2)),
        )
        self.accepted = ['-inf', -100., -3, 2, 6, 300]
        self.rejected = ['nan', '+inf', 7, 3, -0.5, -5]

#------------------------------------------------------------------------------


if __name__ == '__main__':
    main()



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.core.util.xmltools import parse, NameSpace, NameSpaceMap
from eoxserver.core.util.iteratortools import pairwise
from eoxserver.core.decoders import xml, to_dict, InvalidParameterException
from eoxserver.core import Component

NS_EOP_20 = NameSpace("http://www.opengis.net/eop/2.0", "eop")
NS_OPT_20 = NameSpace("http://www.opengis.net/opt/2.0", "opt")
NS_SAR_20 = NameSpace("http://www.opengis.net/sar/2.0", "sar")
NS_ATM_20 = NameSpace("http://www.opengis.net/atm/2.0", "atm")

namespaces_20 = [NS_EOP_20, NS_OPT_20, NS_SAR_20, NS_ATM_20]

NS_EOP_21 = NameSpace("http://www.opengis.net/eop/2.1", "eop")
NS_OPT_21 = NameSpace("http://www.opengis.net/opt/2.1", "opt")
NS_SAR_21 = NameSpace("http://www.opengis.net/sar/2.1", "sar")
NS_ATM_21 = NameSpace("http://www.opengis.net/atm/2.1", "atm")

namespaces_21 = [NS_EOP_21, NS_OPT_21, NS_SAR_21, NS_ATM_21]

NS_OM = NameSpace("http://www.opengis.net/om/2.0", "om")
NS_GML = NameSpace("http://www.opengis.net/gml/3.2", "gml")

nsmap_20 = NameSpaceMap(NS_GML, NS_OM, *namespaces_20)
nsmap_21 = NameSpaceMap(NS_GML, NS_OM, *namespaces_21)
nsmap_gml = NameSpaceMap(NS_GML)


class EOOMFormatReader(Component):
    def test(self, obj):
        tree = parse(obj)
        tag = tree.getroot().tag if tree is not None else None
        return tree is not None and tag in [
            ns('EarthObservation') for ns in namespaces_20 + namespaces_21
        ]

    def read(self, obj):
        tree = parse(obj)
        if tree is not None:
            root = tree.getroot()
            root_ns = root.nsmap[root.prefix]
            use_21 = root_ns in namespaces_21
            decoder = EOOMFormatDecoder(tree, use_21)

            metadata_type = None

            if root_ns in (NS_OPT_20, NS_OPT_21):
                metadata_type = "OPT"
            # TODO: fixme
            # elif root_ns in (NS_ALT_20, NS_ALT_21):
            #     metadata_type = "ALT"
            elif root_ns in (NS_SAR_20, NS_SAR_21):
                metadata_type = "SAR"

            return {
                "identifier": decoder.identifier,
                "begin_time": decoder.begin_time,
                "end_time": decoder.end_time,
                "footprint": MultiPolygon(*decoder.polygons),
                "format": "eogml",
                "metadata": to_dict(EOOMExtraMetadataDecoder(tree, use_21)),
                # "product_metadata": to_dict(
                #     EOOMProductMetadataDecoder(tree, use_21)
                # ),
                # "metadata_type": metadata_type
            }

        raise Exception("Could not parse from obj '%s'." % repr(obj))


def parse_polygon_xml(elem):
    return Polygon(
        parse_ring(
            elem.xpath(
                "gml:exterior/gml:LinearRing/gml:posList", namespaces=nsmap_gml
            )[0].text
        ),
        *map(
            lambda e: parse_ring(e.text),
            elem.xpath(
                "gml:interior/gml:LinearRing/gml:posList", namespaces=nsmap_gml
            )
        )
    )


def parse_ring(string):
    raw_coords = map(float, string.split(" "))
    return [(lon, lat) for lat, lon in pairwise(raw_coords)]


class EOOMNamespaceMixIn(xml.Decoder):
    def __init__(self, tree, use_21):
        if use_21:
            self.namespaces = nsmap_21
        else:
            self.namespaces = nsmap_20
        super(EOOMNamespaceMixIn, self).__init__(tree)


class EOOMFormatDecoder(EOOMNamespaceMixIn, xml.Decoder):
    identifier = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:identifier/text()", type=str, num=1)
    begin_time = xml.Parameter("om:phenomenonTime/gml:TimePeriod/gml:beginPosition/text()", type=parse_iso8601, num=1)
    end_time = xml.Parameter("om:phenomenonTime/gml:TimePeriod/gml:endPosition/text()", type=parse_iso8601, num=1)
    polygons = xml.Parameter("om:featureOfInterest/eop:Footprint/eop:multiExtentOf/gml:MultiSurface/gml:surfaceMember/gml:Polygon", type=parse_polygon_xml, num="+")


class EOOMCollectionMetadataDecoder(EOOMNamespaceMixIn, xml.Decoder):
    spectral_range = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/ eop:wavelengthInformation/eop:WavelengthInformation/eop:spectralRange/text()", type=str, num="?")
    wavelengths = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/ eop:wavelengthInformation/eop:WavelengthInformation/eop:discreteWavelengths/text()", type=str, num="?")
    platform = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:platform/eop:Platform/eop:shortName/text()", type=str, num="?")
    platform_serial_identifier = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:platform/eop:Platform/eop:serialIdentifier/text()", type=str, num="?")
    instrument = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:instrument/eop:Instrument/eop:shortName/text()", type=str, num="?")
    sensor_type = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:sensorType/text()", type=str, num="?")
    composite_type = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:compositeType/text()", type=str, num="?")
    processing_level = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:processingLevel/text()", type=str, num="?")
    orbit_type = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:platform/eop:Platform/eop:orbitType/text()", type=str, num="?")


class EOOMProductMetadataDecoder(EOOMNamespaceMixIn, xml.Decoder):
    parent_identifier = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:parentIdentifier/text()", type=str, num="?")

    production_status = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:status/text()", type=str, num="?")
    acquisition_type = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:acquisitionType/text()", type=str, num="?")

    orbit_number = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:orbitNumber/text()", type=str, num="?")
    orbit_direction = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:orbitDirection/text()", type=str, num="?")

    track = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:wrsLongitudeGrid/text()", type=str, num="?")
    frame = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:wrsLatitudeGrid/text()", type=str, num="?")
    swath_identifier = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:swathIdentifier/text()", type=str, num="?")

    product_version = xml.Parameter("om:result/eop:EarthObservationResult/eop:product/eop:ProductInformation/eop:version/text()", type=str, num="?")
    product_quality_status = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:productQualityDegradation/text()", type=str, num="?")
    product_quality_degradation_tag = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:productQualityDegradationTag/text()", type=str, num="?")
    processor_name = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:processorName/text()", type=str, num="?")
    processing_center = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetadata/eop:processing/eop:ProcessingInformation/eop:processingCenter/text()", type=str, num="?")
    processing_date = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetadata/eop:processing/eop:ProcessingInformation/eop:processingDate/text()", type=parse_iso8601, num="?")
    sensor_mode = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:operationalMode/text()", type=str, num="?")
    archiving_center = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:archivedIn/eop:ArchivingInformation/eop:archivingCenter/text()", type=str, num="?")
    processing_mode = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:ProcessingMode/text()", type=str, num="?")
    creation_date = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:creationDate/text()", type=parse_iso8601, num="?")


class EOOMExtraMetadataDecoder(EOOMNamespaceMixIn, xml.Decoder):
    modification_date = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:modificationDate/text()", type=parse_iso8601, num="?")

    # TODO: get this into models
    # resolution = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:resolution/text()", type=str, num="?")

    cloud_cover = xml.Parameter("om:result/opt:EarthObservationResult/opt:cloudCoverPercentage/text()|om:result/atm:EarthObservationResult/atm:cloudCoverPercentage/text()", type=float, num="?")
    snow_cover = xml.Parameter("om:result/opt:EarthObservationResult/opt:snowCoverPercentage/text()|om:result/atm:EarthObservationResult/atm:snowCoverPercentage/text()", type=float, num="?")
    lowest_location = xml.Parameter("atm:EarthObservation/om:resultOf/atm:EarthObservationResult/atm:dataLayers/atm:DataLayer/atm:lowestLocation/text()", type=float, num="?")
    highest_location = xml.Parameter("atm:EarthObservation/om:resultOf/atm:EarthObservationResult/atm:dataLayers/atm:DataLayer/atm:highestLocation/text()", type=float, num="?")

    acquisition_station = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:downlinkedTo/eop:DownlinkInformation/eop:acquisitionStation/text()", type=str, num="?")
    availability_time = xml.Parameter("om:resultTime/gml:TimeInstant/gml:timePosition/text()", type=parse_iso8601, num="?")
    acquisition_sub_type = xml.Parameter("eop:metaDataProperty/eop:EarthObservationMetaData/eop:acquisitionSubType/text()", type=str, num="?")
    start_time_from_ascending_node = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:startTimeFromAscendingNode/text()", type=int, num="?")
    completion_time_from_ascending_node = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:completionTimeFromAscendingNode/text()", type=int, num="?")

    illumination_azimuth_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:illuminationAzimuthAngle/text()", type=float, num="?")
    illumination_zenith_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:illuminationZenithAngle/text()", type=float, num="?")
    illumination_elevation_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:illuminationElevationAngle/text()", type=float, num="?")

    polarisation_mode = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:polarisationMode/text()", type=str, num="?")
    polarization_channels = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:polarisationChannels/text()", type=str, num="?")
    antenna_look_direction = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:antennaLookDirection/text()", type=str, num="?")
    minimum_incidence_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:minimumIncidenceAngle/text()", type=float, num="?")
    maximum_incidence_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:maximumIncidenceAngle/text()", type=float, num="?")
    across_track_incidence_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:acrossTrackIncidenceAngle/text()", type=float, num="?")
    along_track_incidence_angle = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:alongTrackIncidenceAngle/text()", type=float, num="?")
    doppler_frequency = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:dopplerFrequency/text()", type=float, num="?")
    incidence_angle_variation = xml.Parameter("om:procedure/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:incidenceAngleVariation/text()", type=float, num="?")



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.contrib.vsi import open as vsi_open

from eoxserver.resources.coverages.metadata.product_formats import get_readers


class ProductMetadataComponent(object):
    def read_product_metadata_file(self, path):
        for reader_cls in get_readers():
            reader = reader_cls()

            if hasattr(reader, 'test_path') and reader.test_path(path):
                return reader.read_path(path)
            elif hasattr(reader, 'test'):
                try:
                    f = vsi_open(path)
                except IOError:
                    continue

                if reader.test(f):
                    f.seek(0)
                    return reader.read(f)

        if f:
            f.close()

        return {}

    def collect_package_metadata(self, storage, handler, cache=None):
        path = handler.get_vsi_path(storage.url)
        for reader_cls in get_readers():
            reader = reader_cls()
            if hasattr(reader, 'test_path'):
                if reader.test_path(path):
                    return reader.read_path(path)
            else:
                try:
                    with open(path) as f:
                        if hasattr(reader, 'test') and f and reader.test(f):
                            return reader.read(f)
                except IOError:
                    pass

        raise Exception('No suitable metadata reader found.')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-06 15:13
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('backends', '0002_storage_auth'),
    ]

    operations = [
        migrations.AlterField(
            model_name='storage',
            name='name',
            field=models.CharField(blank=True, max_length=1024, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='storageauth',
            name='name',
            field=models.CharField(blank=True, max_length=1024, null=True, unique=True),
        ),
    ]



#-------------------------------------------------------------------------------
#
#  WPS input and output parameters and data types
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from .base import Parameter
from .literaldata import LiteralData
from .complexdata import (
    ComplexData, CDBase, CDObject, CDTextBuffer, CDByteBuffer,
    CDAsciiTextBuffer, CDFileWrapper, CDFile, CDPermanentFile,
)
from .formats import (
    Format, FormatText, FormatXML, FormatJSON,
    FormatBinaryRaw, FormatBinaryBase64,
)
from .codecs import Codec, CodecBase64, CodecRaw
from .bboxdata import BoundingBox, BoundingBoxData
from .units import UnitOfMeasure, UnitLinear
from .allowed_values import (
    BaseAllowed, AllowedAny, AllowedEnum, AllowedRange,
    AllowedRangeCollection, AllowedByReference
)
from .data_types import (
    DTYPES, BaseType, Boolean, Integer, Double, String,
    Duration, Date, Time, DateTime, DateTimeTZAware
)
from .crs import CRSType
from .inputs import InputReference, InputData
from .response_form import (
    Output, ResponseForm, ResponseDocument, RawDataOutput
)

class RequestParameter(object):
    """ Special input parameter extracting input from the request metadata.
    This might be used to pass information such as, e.g., HTTP headers or
    user authentication to the process like a regular input variable.

    This class is the base class and it expected that `parse_request`
    method get overloaded by inheritance or by a function passed as
    an argument to the constructor.
    """
    # pylint: disable=method-hidden, too-few-public-methods

    def __init__(self, request_parser=None):
        if request_parser:
            self.parse_request = request_parser

    def parse_request(self, request):
        """ Method extracting information from the Django HTTP request object.
        """
        raise NotImplementedError


def fix_parameter(name, prm):
    """ Expand short-hand definition of the parameter."""
    if isinstance(prm, Parameter):
        return prm
    elif isinstance(prm, RequestParameter):
        # The leading backslash indicates an internal parameter.
        # Note: backslash is not an allowed URI character and it cannot appear
        # in the WPS inputs' names.
        prm.identifier = "\\" + name
        return prm
    else:
        return LiteralData(name, dtype=prm)


class Reference(object):
    """ Output reference. An instance of this class defines a CommplexData
    output passed by a reference. The output must be stored in a file.

    Constructor parameters:
        path        path to the output file in the local file-system
        href        public URL of the output reference
        mime_type   output ComplexData mime-type
        encoding    output ComplexData encoding
        schema      output ComplexData schema
    """
    # pylint: disable=too-few-public-methods, too-many-arguments
    def __init__(self, path, href, mime_type=None, encoding=None, schema=None,
                 **kwarg):
        self.path = path
        self.href = href
        self.mime_type = mime_type
        self.encoding = encoding
        self.schema = schema



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from lxml.builder import ElementMaker
from lxml.etree import ProcessingInstruction, ElementTree
from django.conf import settings
from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap


ns_xlink = NameSpace("http://www.w3.org/1999/xlink", "xlink")
ns_ows = NameSpace("http://www.opengis.net/ows/1.1", "ows", "http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd")
ns_xml = NameSpace("http://www.w3.org/XML/1998/namespace", "xml")

nsmap = NameSpaceMap(ns_ows)
OWS = ElementMaker(namespace=ns_ows.uri, nsmap=nsmap)


class OWS11ExceptionXMLEncoder(XMLEncoder):
    def encode_exception(self, message, version, code, locator=None):
        exception_attributes = {
            "exceptionCode": code
        }

        if locator:
            exception_attributes["locator"] = locator

        exception_text = (OWS("ExceptionText", message),) if message else ()

        xml_tree = ElementTree(OWS("ExceptionReport",
            OWS("Exception", *exception_text, **exception_attributes
            ), version=version, **{ns_xml("lang"): "en"}
        ))

        xls_url = getattr(settings, "OWS11_EXCEPTION_XSL", None)
        if xls_url:
            xml_tree.getroot().addprevious(ProcessingInstruction(
                'xml-stylesheet', 'type="text/xsl" href="%s"' % xls_url
            ))

        return xml_tree

    def get_schema_locations(self):
        return nsmap.schema_locations



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


import uuid

from django.http import QueryDict
from lxml.builder import ElementMaker
try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse
from django.utils.http import urlencode

from eoxserver.contrib import ogr, vsi
from eoxserver.core.util.timetools import isoformat
from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.resources.coverages import models
from eoxserver.services.gml.v32.encoders import GML32Encoder


class BaseOGRResultFormat(object):
    """ Base ckass for result formats using OGR for encoding the records.
    """
    abstract = True

    driver_name = None
    extension = None

    def encode(self, request, collection_id, queryset, search_context):
        """ Encode a query set as an OGR datasource and retrieve its contents.
        """
        # create a datasource and its fields
        driver = self.get_driver()
        filename = self.get_filename()
        ds = self.create_datasource(driver, filename)
        layer = self.create_layer(ds)
        self.create_fields(layer)
        definition = layer.GetLayerDefn()

        # encode the objects
        for eo_object in queryset:
            feature = self.create_feature(layer, definition)
            self.set_feature_values(feature, eo_object)
            layer.CreateFeature(feature)

        # close datasource and read contents
        ds.Destroy()

        with vsi.open(filename) as f:
            content = f.read()

        # perform cleanup and return content + mimetype
        self.cleanup(driver, ds, filename)
        return content

    def get_driver(self):
        """ Get the OGR driver.
        """
        return ogr.GetDriverByName(self.driver_name)

    def get_filename(self):
        """ Get the filename for the temporary file.
        """
        return "/vsimem/%s%s" % (uuid.uuid4().hex, self.extension)

    def create_datasource(self, driver, filename):
        """ Create the OGR DataSource. This needs to be overriden in formats
            that require certain creation options.
        """
        return driver.CreateDataSource(filename)

    def create_layer(self, datasource):
        """ Create the layer in the DataSource.
        """
        return datasource.CreateLayer("layer", geom_type=ogr.wkbMultiPolygon)

    def create_fields(self, layer):
        """ Create the field definitions of the layer. By default, it contains
            definitions of the `id`, `begin_time` and `end_time` fields. For
            certain formats, this needs to be overridden.
        """
        layer.CreateField(ogr.FieldDefn("id", ogr.OFTString))
        layer.CreateField(ogr.FieldDefn("begin_time", ogr.OFTString))
        layer.CreateField(ogr.FieldDefn("end_time", ogr.OFTString))

    def create_feature(self, layer, definition):
        """ Create a feature from the given definition.
        """
        return ogr.Feature(definition)

    def set_feature_values(self, feature, eo_object):
        """ Set the values and the geometry of the feature. This needs to be
            inline with the :meth:`create_fields` method.
        """
        feature.SetGeometry(
            ogr.CreateGeometryFromWkb(str(eo_object.footprint.wkb))
        )
        feature.SetField("id", eo_object.identifier.encode("utf-8"))
        if eo_object.begin_time:
            feature.SetField("begin_time", isoformat(eo_object.begin_time))
        if eo_object.end_time:
            feature.SetField("end_time", isoformat(eo_object.end_time))

    def cleanup(self, driver, datasource, filename):
        """ Perform any necessary cleanup steps, like removing the temporary
            file.
        """
        driver.DeleteDataSource(filename)


ns_atom = NameSpace("http://www.w3.org/2005/Atom", "atom")
ns_opensearch = NameSpace("http://a9.com/-/spec/opensearch/1.1/", "opensearch")
ns_dc = NameSpace("http://purl.org/dc/elements/1.1/", "dc")
ns_georss = NameSpace("http://www.georss.org/georss", "georss")
ns_media = NameSpace("http://search.yahoo.com/mrss/", "media")
ns_owc = NameSpace("http://www.opengis.net/owc/1.0", "owc")

nsmap = NameSpaceMap(ns_atom, ns_dc, ns_georss, ns_media, ns_owc)

ATOM = ElementMaker(namespace=ns_atom.uri)
OS = ElementMaker(namespace=ns_opensearch.uri)
DC = ElementMaker(namespace=ns_dc.uri, nsmap=nsmap)
GEORSS = ElementMaker(namespace=ns_georss.uri, nsmap=nsmap)
MEDIA = ElementMaker(namespace=ns_media.uri, nsmap=nsmap)
OWC = ElementMaker(namespace=ns_owc.uri, nsmap=nsmap)


class BaseFeedResultFormat(object):
    """ Abstract base component for feed result formats like RSS and atom. Adds
    functionality to encode the paging mechanisms by using ``atom:link``s.
    """
    abstract = True

    def encode_feed_links(self, request, search_context):
        sc = search_context
        qdict = QueryDict(request.GET.urlencode(), mutable=True)
        qdict.pop("startIndex", None)
        links = [
            ATOM("link",
                rel="search", type="application/opensearchdescription+xml",
                href=request.build_absolute_uri(
                    reverse("opensearch:description")
                )
            ),
            ATOM("link",
                rel="self", type=self.mimetype,
                href=request.build_absolute_uri()
            ),
            ATOM("link",
                rel="first", type=self.mimetype,
                href="%s?%s" % (
                    request.build_absolute_uri(request.path),
                    qdict.urlencode()
                )
            )
        ]

        # add link to last page
        qdict["startIndex"] = str(sc.page_count * (sc.page_size or sc.count))
        links.append(
            ATOM("link",
                rel="last", type=self.mimetype,
                href="%s?%s" % (
                    request.build_absolute_uri(request.path),
                    qdict.urlencode()
                )
            )
        )

        # if not already on the first page, include link to previous page
        if sc.current_page != 0:
            qdict["startIndex"] = str(max(0, sc.start_index - sc.page_size))
            links.append(
                ATOM("link",
                    rel="previous", type=self.mimetype,
                    href="%s?%s" % (
                        request.build_absolute_uri(request.path),
                        qdict.urlencode()
                    )
                )
            )

        # if not already on the last page, include link to next page
        if sc.start_index + sc.count < sc.total_count:
            qdict["startIndex"] = str(sc.start_index + sc.count)
            links.append(
                ATOM("link",
                    rel="next", type=self.mimetype,
                    href="%s?%s" % (
                        request.build_absolute_uri(request.path),
                        qdict.urlencode()
                    )
                )
            )
        return links

    def encode_opensearch_elements(self, search_context):
        return [
            OS("totalResults", str(search_context.total_count)),
            OS("startIndex", str(search_context.start_index or 0)),
            OS("itemsPerPage", str(search_context.count)),
            OS("Query", role="request", **dict(
                ("{%s}%s" % (search_context.namespaces[prefix], name), value)
                for prefix, params in search_context.parameters.items()
                for name, value in params.items()
            ))
        ]

    def encode_item_links(self, request, collection_id, item):
        links = []
        if isinstance(item, models.Collection):
            # add link to opensearch collection search
            links.append(
                ATOM("link",
                    rel="search", type="application/opensearchdescription+xml",
                    href=request.build_absolute_uri(
                        reverse("opensearch:collection:description", kwargs={
                            'collection_id': item.identifier
                        })
                    )
                )
            )
            # TODO: link to WMS (GetCapabilities)

        if isinstance(item, models.Product):
            footprint = item.footprint
            if footprint:

                links.append(
                    ATOM("link", rel="enclosure",
                        href=self._create_download_link(request, item)
                    )
                )

                wms_get_capabilities = self._create_wms_capabilities_link(
                    request, item
                )

                thumbnail_link = self._create_thumbail_link(request, item)
                wms_small = self._create_map_link(request, item, 100)
                wms_large = self._create_map_link(request, item, 500)

                # media RSS style links
                if wms_large:
                    # "Browse" image
                    links.append(
                        MEDIA("content",
                            MEDIA("category", "QUICKLOOK"),
                            url=wms_large
                        )
                    )

                if wms_small:
                    # "Thumbnail" image
                    links.append(
                        MEDIA("content",
                            MEDIA("category", "THUMBNAIL"),
                            url=thumbnail_link or wms_small
                        )
                    )

                links.extend([
                    OWC("offering",
                        OWC("operation",
                            code="GetCapabilities", method="GET",
                            type="application/xml", href=wms_get_capabilities
                        ),
                        OWC("operation",
                            code="GetMap", method="GET",
                            type="image/png", href=wms_large
                        ),
                        code="http://www.opengis.net/spec/owc-atom/1.0/req/wms",
                    ),
                ])

                wcs_offering = OWC("offering",
                    OWC("operation",
                        code="GetCapabilities", method="GET",
                        type="application/xml",
                        href=request.build_absolute_uri(
                            "%s?service=WCS&version=2.0.1"
                            "&request=GetCapabilities"
                            % reverse("ows")
                        )
                    ),
                    code="http://www.opengis.net/spec/owc-atom/1.0/req/wcs",
                )
                for coverage in item.coverages.all():
                    wcs_offering.extend(self.encode_coverage_offerings(
                        request, coverage
                    ))

                links.append(wcs_offering)

        if isinstance(item, models.Coverage):
            # add a link for a Describe and GetCoverage request for
            # metadata and data download

            wcs_get_capabilities = request.build_absolute_uri(
                "%s?service=WCS&version=2.0.1&request=GetCapabilities"
                % reverse("ows")
            )

            links.extend([
                ATOM("link", rel="enclosure",
                    href=self._create_coverage_link(
                        request, item
                    )
                ),
                ATOM("link", rel="via",
                    href=self._create_coverage_description_link(
                        request, item
                    )
                ),
                # "Browse" image
                # ATOM("link", rel="icon", href=wms_large),
            ])

            # OWC offerings for WCS
            links.extend([
                OWC("offering",
                    OWC("operation",
                        code="GetCapabilities", method="GET",
                        type="application/xml", href=wcs_get_capabilities
                    ),
                    *self.encode_coverage_offerings(request, item),
                    **{
                        "code": "http://www.opengis.net/spec/owc-atom/1.0/req/wcs"
                    }
                )
            ])

        semantic_to_rel = {
            1: 'alternate',
            2: 'describedby',
        }

        links.extend([
            ATOM("link",
                rel=semantic_to_rel[metadata_item.semantic],
                href=self._make_metadata_href(request, item, metadata_item)
            )
            for metadata_item in item.metadata_items.filter(
                semantic__in=semantic_to_rel.keys()
            )
        ])

        return links

    def encode_summary(self, request, collection_id, item):
        pass

    def encode_coverage_offerings(self, request, coverage):
        return [
            OWC("operation",
                code="DescribeCoverage", method="GET",
                type="application/xml",
                href=self._create_coverage_description_link(request, coverage)
            ),
            OWC("operation",
                code="GetCoverage", method="GET",
                type="image/tiff", href=self._create_coverage_link(
                    request, coverage
                )
            )
        ]

    def encode_spatio_temporal(self, item):
        entries = []

        begin_time = item.begin_time
        end_time = item.end_time
        if begin_time and end_time:
            if begin_time != end_time:
                entries.append(
                    DC("date", "%s/%s" % (
                        isoformat(begin_time), isoformat(end_time)
                    ))
                )
            else:
                entries.append(DC("date", isoformat(begin_time)))

        if item.footprint:
            extent = item.footprint.extent
            entries.append(
                GEORSS("box",
                    "%f %f %f %f" % (
                        extent[1], extent[0], extent[3], extent[2]
                    )
                )
            )
            entries.append(
                GEORSS("where",
                    GML32Encoder().encode_multi_surface(
                        item.footprint, item.identifier
                    )
                )
            )

        return entries

    def _create_wms_capabilities_link(self, request, item):
        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="WMS",
                    request="GetCapabilities",
                    cql="identifier='%s'" % item.identifier,
                ))
            )
        )

    def _create_map_link(self, request, item, size):
        footprint = item.footprint

        if footprint:
            minx, miny, maxx, maxy = footprint.extent

            fx = 1.0
            fy = 1.0

            # in case of points
            if minx == maxx or miny == maxy:
                return None

            if (maxx - minx) > (maxy - miny):
                fy = (maxy - miny) / (maxx - minx)
            else:
                fx = (maxx - minx) / (maxy - miny)

            return request.build_absolute_uri(
                "%s?%s" % (
                    reverse("ows"), urlencode(dict(
                        service="WMS",
                        version="1.3.0",
                        request="GetMap",
                        layers=item.identifier,
                        format="image/png",
                        TRANSPARENT="true",
                        width=int(size * fx),
                        height=int(size * fy),
                        CRS="EPSG:4326",
                        STYLES="",
                        BBOX="%f,%f,%f,%f" % (miny, minx, maxy, maxx)
                    ))
                )
            )
        return None

    def _create_coverage_link(self, request, coverage):
        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="WCS",
                    version="2.0.1",
                    request="GetCoverage",
                    coverageId=coverage.identifier,
                ))
            )
        )

    def _create_coverage_description_link(self, request, coverage):
        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="WCS",
                    version="2.0.1",
                    request="DescribeCoverage",
                    coverageId=coverage.identifier,
                ))
            )
        )

    def _create_eo_coverage_set_description(self, request, eo_object):
        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="WCS",
                    version="2.0.1",
                    request="DescribeEOCoverageSet",
                    eoId=eo_object.identifier,
                ))
            )
        )

    def _create_self_link(self, request, collection_id, item, format=None):
        if collection_id is None:
            return request.build_absolute_uri(
                "%s?%s" % (
                    reverse("opensearch:search", kwargs={
                        "format_name": format if format else self.name
                    }),
                    urlencode(dict(
                        uid=item.identifier
                    ))
                )
            )

        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("opensearch:collection:search", kwargs={
                    "collection_id": collection_id,
                    "format_name": format if format else self.name
                }),
                urlencode(dict(
                    uid=item.identifier
                ))
            )
        )

    def _create_download_link(self, request, product):
        package = product.package
        if package:
            if package.storage_type in ('HTTP', 'FTP'):
                return package.url

        return request.build_absolute_uri(
            "%s?%s" % (
                reverse("ows"), urlencode(dict(
                    service="DSEO",
                    version="1.0.0",
                    request="GetProduct",
                    ProductURI=product.identifier,
                ))
            )
        )

    def _create_thumbail_link(self, request, item):
        semantic = models.MetaDataItem.semantic_codes['thumbnail']
        if item.metadata_items.filter(semantic=semantic).exists():
            return request.build_absolute_uri(
                reverse("metadata", kwargs={
                    'identifier': item.identifier,
                    'semantic': 'thumbnail'
                })
            )

    def _make_metadata_href(self, request, item, metadata_item):
        semantic_name = models.MetaDataItem.semantic_names[metadata_item.semantic]
        return request.build_absolute_uri(
            reverse("metadata", kwargs={
                'identifier': item.identifier,
                'semantic': semantic_name
            })
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module contains facilities to help decoding configuration files.
It relies on the :mod:`ConfigParser` module for actually reading the file.
"""

import sys

try:
    from ConfigParser import NoOptionError, NoSectionError
except ImportError:
    from configparser import NoOptionError, NoSectionError

from django.utils.six import with_metaclass


def section(name):
    """ Helper to set the section of a :class:`Reader`.
    """
    frame = sys._getframe(1)
    locals_ = frame.f_locals

    # Some sanity checks
    assert locals_ is not frame.f_globals and '__module__' in locals_, \
        'implements() can only be used in a class definition'

    locals_["section"] = name


class Option(property):
    """ The :class:`Option` is used as a :class:`property` for :class:`Reader`
        subclasses.

        :param key: the lookup key; defaults to the property name of the
                    :class:`Reader`.
        :param type: the type to parse the raw value; by default the raw
                     string is returned
        :param separator: the separator for list options; by default no list
                          is assumed
        :param required: if ``True`` raise an error if the option does not
                         exist
        :param default: the default value
        :param section: override the section for this option
    """

    def __init__(self, key=None, type=None, separator=None, required=False,
                 default=None, section=None, doc=None):

        super(Option, self).__init__(self.fget)

        self.key = key  # needs to be set by the reader metaclass
        self.type = type
        self.separator = separator
        self.required = required
        self.default = default

        if section is None:
            frame = sys._getframe(1)
            section = frame.f_locals.get("section")

        self.section = section

    def fget(self, reader):
        
        section = self.section or reader.section
        try:
            if self.type is bool:
                raw_value = reader._config.getboolean(section, self.key)
            else:
                raw_value = reader._config.get(section, self.key)
        except (NoOptionError, NoSectionError) as e:
            if not self.required:
                return self.default
            raise e

        if self.separator is not None:
            return map(self.type, raw_value.split(self.separator))

        elif self.type:
            return self.type(raw_value)

        else:
            return raw_value

    def check(self, reader):
        # TODO: perform checking of config
        #  - required option?
        #  - can parse type?
        pass


class ReaderMetaclass(type):
    def __init__(cls, name, bases, dct):
        for key, value in dct.items():
            if isinstance(value, Option) and value.key is None:
                value.key = key

                value.__doc__ = "%s.%s" % (value.section, value.key)

        super(ReaderMetaclass, cls).__init__(name, bases, dct)


class Reader(with_metaclass(ReaderMetaclass, object)):
    """ Base class for config readers.

    :param config: an instance of :class:`ConfigParser.RawConfigParser`

    Readers should be used as such:
    ::

        from ConfigParser import RawConfigParser
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import StringIO
        from textwrap import dedent
        from eoxserver.core.decoders import config

        class ExampleReader(config.Reader):
            section = "example_section"
            string_opt = config.Option()
            string_list_opt = config.Option(separator=",")
            integer_opt = config.Option(type=int)

            section = "other_section"
            mandatory_opt = config.Option(required=True)
            optional_opt = config.Option(default="some_default")

            special_section_opt = config.Option(section="special_section")

        f = StringIO(dedent('''
            [example_section]
            string_opt = mystring
            string_list_opt = my,string,list
            integer_opt = 123456
            [other_section]
            mandatory_opt = mandatory_value
            # optional_opt = no value

            [special_section]
            special_section_opt = special_value
        '''))

        parser = RawConfigParser()
        parser.readfp(f)
        reader = ExampleReader(parser)

        print reader.string_opt
        print reader.string_list_opt
        print reader.integer_opt
        print reader.mandatory_opt
        print reader.optional_opt
        ...
    """

    __metaclass__ = ReaderMetaclass

    section = None

    def __init__(self, config):
        self._config = config



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class ServiceHandlerInterface(object):
    """ Interface for OWS Service handlers.
    """

    @property
    def service(self):
        """ The name of the supported service in uppercase letters. This can 
            also be an iterable, if the handler shall support more than one 
            service specifier.
            Some service specifications demand that the service parameter can be
            omitted for certain requests. In this case this property can alse be
            ``None`` or contain ``None``.
        """

    @property
    def versions(self):
        """ An iterable of all supported versions as strings.
        """

    @property
    def request(self):
        """ The supported request method.
        """

    def handle(self, request):
        """ The main handling method. Takes a `django.http.Request` object as 
            single parameter.
        """

    @property
    def constraints(self):
        """ Optional property to return a dict with constraints for default 
            values.
        """

    @property
    def index(self):
        """ Optional. The index this service handler shall have when being 
            reported in a capabilities document.
        """


class ExceptionHandlerInterface(object): 
    """ Interface for OWS exception handlers.
    """

    @property
    def service(self):
        """ The name of the supported service in uppercase letters. This can 
            also be an iterable, if the handler shall support more than one 
            service specifier.
            Some service specifications demand that the service parameter can be
            omitted for certain requests. In this case this property can alse be
            ``None`` or contain ``None``.
        """

    @property
    def versions(self):
        """ An iterable of all supported versions as strings.
        """

    @property
    def request(self):
        """ The supported request method.
        """

    def handle_exception(self, request, exception):
        """ The main exception handling method. Parameters are an object of the 
            `django.http.Request` type and the raised exception.
        """


class GetServiceHandlerInterface(ServiceHandlerInterface):
    """ Interface for service handlers that support HTTP GET requests.
    """


class PostServiceHandlerInterface(ServiceHandlerInterface):
    """ Interface for service handlers that support HTTP POST requests.
    """


class VersionNegotiationInterface(ServiceHandlerInterface):
    """ Interface for handlers that contribute to version negotiation.
    """



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from lxml import etree
from lxml.builder import E

from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.util.xmltools import parse
from eoxserver.core.util.timetools import isoformat
from eoxserver.core.util.iteratortools import pairwise
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.core.decoders import xml
from eoxserver.core import Component

class NativeFormat(Component):
    formats = ("native", )

    def test(self, obj):
        xml = parse(obj)
        return xml is not None and xml.getroot().tag == "Metadata"

    def get_format_name(self, obj):
        return "native"

    def read(self, obj):
        tree = parse(obj)
        if tree is not None:
            decoder = NativeFormatDecoder(tree)
            return {
                "identifier": decoder.identifier,
                "begin_time": decoder.begin_time,
                "end_time": decoder.end_time,
                "footprint": MultiPolygon(*decoder.polygons),
                "format": "native"
            }
        raise Exception("Could not parse from obj '%s'." % repr(obj))

    def write(self, values, file_obj, format=None, encoding=None, pretty=False):
        def flip(point):
            return point[1], point[0]

        # ignore format
        tree = E.Metadata(
            E.EOID(values["identifier"]),
            E.BeginTime(isoformat(values["begin_time"])),
            E.EndTime(isoformat(values["end_time"])),
            E.Footprint(
                *map(lambda polygon:
                    E.Polygon(
                        E.Exterior(
                            " ".join([
                                "%f %f" % flip(point)
                                for point in polygon.exterior_ring
                            ])
                        ),
                        *[E.Interior(
                            " ".join([
                                "%f %f" % flip(point)
                                for point in interior
                            ])
                        ) for interior in polygon[1:]]
                    ),
                    values["footprint"]
                )
            )
        )

        file_obj.write(
            etree.tostring(tree, pretty_print=pretty, encoding=encoding)
        )


def parse_polygon_xml(elem):
    return Polygon(
        parse_ring(elem.findtext("Exterior")),
        *map(lambda e: parse_ring(e.text), elem.findall("Interior"))
    )


def parse_ring(string):
    raw_coords = map(float, string.split(" "))
    return [(lon, lat) for lat, lon in pairwise(raw_coords)]


class NativeFormatDecoder(xml.Decoder):
    identifier = xml.Parameter("EOID/text()")
    begin_time = xml.Parameter("BeginTime/text()", type=parse_iso8601)
    end_time = xml.Parameter("EndTime/text()", type=parse_iso8601)
    polygons = xml.Parameter("Footprint/Polygon", type=parse_polygon_xml, num="+")



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from django.contrib.gis.db import models


class Castable(models.Model):
    """ Model mix-in for 'castable' types. With this MixIn, type information and
        completed models can be retrieved.
    """

    @property
    def real_type(self):
        # if not saved, use the actual type
        if not self.id:
            return type(self)

        return self.type_registry[self.real_content_type]

    def __init__(self, *args, **kwargs):
        super(Castable, self).__init__(*args, **kwargs)
        if not self.id:
            for type_id, cls in self.type_registry.items():
                if cls == type(self):
                    self.real_content_type = type_id
                    break

    def cast(self, refresh=False):
        """'Cast' the model to its actual type, if it is not already. This
        invokes a database lookup if the real type is not the same as the type
        of the current model.
        """

        # don't perform a cast if not necessary
        real_type = self.real_type
        if real_type == type(self) and not refresh:
            return self

        return self.real_type.objects.get(pk=self.pk)

    class Meta:
        abstract = True



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
import math
from itertools import product
import numpy

from eoxserver.contrib import gdal, gdal_array, osr, ogr
from eoxserver.processing.preprocessing.util import (
    get_limits, create_temp, copy_metadata, copy_projection, cleanup_temp
)
from eoxserver.resources.coverages.crss import (
    parseEPSGCode, fromShortCode, fromURL, fromURN, fromProj4Str
)
from django.utils.six import string_types

logger = logging.getLogger(__name__)

#===============================================================================
# Dataset Optimization steps
#===============================================================================


class DatasetOptimization(object):
    """ Abstract base class for dataset optimization steps. Each optimization
        step shall be callable and return the dataset or a copy thereof if
        necessary.
    """

    def __call__(self, ds):
        raise NotImplementedError


class ReprojectionOptimization(DatasetOptimization):
    """ Dataset optimization step to reproject the dataset into a predefined
        projection identified by an SRID.
    """

    def __init__(self, crs_or_srid, temporary_directory=None):
        if isinstance(crs_or_srid, int):
            pass
        elif isinstance(crs_or_srid, string_types):
            crs_or_srid = parseEPSGCode(crs_or_srid, (fromShortCode, fromURL,
                                                      fromURN, fromProj4Str))
        else:
            raise ValueError("Unable to obtain CRS from '%s'." %
                             type(crs_or_srid).__name__)

        self.srid = crs_or_srid
        self.temporary_directory = temporary_directory

    def __call__(self, src_ds):
        logger.info("Applying ReprojectionOptimization")
        # setup
        src_sr = osr.SpatialReference()
        src_sr.ImportFromWkt(src_ds.GetProjection())

        dst_sr = osr.SpatialReference()
        dst_sr.ImportFromEPSG(self.srid)

        if src_sr.IsSame(dst_sr) and (src_ds.GetGeoTransform()[1] > 0) \
                and (src_ds.GetGeoTransform()[5] < 0) \
                and (src_ds.GetGeoTransform()[2] == 0) \
                and (src_ds.GetGeoTransform()[4] == 0):
            logger.info("Source and destination projection are equal and image "
                        "is not flipped or has rotated axes. Thus, no "
                        "reprojection is required.")
            return src_ds

        # create a temporary dataset to get information about the output size
        tmp_ds = gdal.AutoCreateWarpedVRT(src_ds, None, dst_sr.ExportToWkt(),
                                          gdal.GRA_Bilinear, 0.125)

        try:
            # create the output dataset
            dst_ds = create_temp(tmp_ds.RasterXSize, tmp_ds.RasterYSize,
                                 src_ds.RasterCount,
                                 src_ds.GetRasterBand(1).DataType,
                                 temp_root=self.temporary_directory)

            # initialize with no data
            for i in range(src_ds.RasterCount):
                src_band = src_ds.GetRasterBand(i+1)
                if src_band.GetNoDataValue() is not None:
                    dst_band = dst_ds.GetRasterBand(i+1)
                    dst_band.SetNoDataValue(src_band.GetNoDataValue())
                    dst_band.Fill(src_band.GetNoDataValue())

            # reproject the image
            dst_ds.SetProjection(dst_sr.ExportToWkt())
            dst_ds.SetGeoTransform(tmp_ds.GetGeoTransform())

            gdal.ReprojectImage(src_ds, dst_ds,
                                src_sr.ExportToWkt(),
                                dst_sr.ExportToWkt(),
                                gdal.GRA_Bilinear)

            tmp_ds = None

            # copy the metadata
            copy_metadata(src_ds, dst_ds)

            return dst_ds
        except:
            cleanup_temp(dst_ds)
            raise


class BandSelectionOptimization(DatasetOptimization):
    """ Dataset optimization step which selects a number of bands and their
    respective scale and copies them to the result dataset.
    """

    def __init__(self, bands, datatype=gdal.GDT_Byte, temporary_directory=None):
        # preprocess bands list
        # TODO: improve
        self.bands = map(lambda b: b if len(b) == 3 else (b[0], None, None),
                         bands)
        self.datatype = datatype
        self.temporary_directory = temporary_directory

    def __call__(self, src_ds):
        logger.info("Applying BandSelectionOptimization")
        try:
            dst_ds = create_temp(src_ds.RasterXSize, src_ds.RasterYSize,
                                 len(self.bands), self.datatype,
                                 temp_root=self.temporary_directory)
            dst_range = get_limits(self.datatype)

            multiple, multiple_written = 0, False

            for dst_index, (src_index, dmin, dmax) in enumerate(self.bands, 1):
                # check if next band is equal
                if dst_index < len(self.bands) and \
                        (src_index, dmin, dmax) == self.bands[dst_index]:
                    multiple += 1
                    continue
                # check that src band is available
                if src_index > src_ds.RasterCount:
                    continue

                # initialize with zeros if band is 0
                if src_index == 0:
                    src_band = src_ds.GetRasterBand(1)
                    data = numpy.zeros(
                        (src_band.YSize, src_band.XSize),
                        dtype=gdal_array.codes[self.datatype]
                    )
                    src_min, src_max = (0, 0)
                # use src_ds band otherwise
                else:
                    src_band = src_ds.GetRasterBand(src_index)
                    src_min, src_max = src_band.ComputeRasterMinMax()

                # get min/max values or calculate from band
                if dmin is None:
                    dmin = get_limits(src_band.DataType)[0]
                elif dmin == "min":
                    dmin = src_min
                if dmax is None:
                    dmax = get_limits(src_band.DataType)[1]
                elif dmax == "max":
                    dmax = src_max
                src_range = (float(dmin), float(dmax))

                block_x_size, block_y_size = src_band.GetBlockSize()

                num_x = int(math.ceil(float(src_band.XSize) / block_x_size))
                num_y = int(math.ceil(float(src_band.YSize) / block_y_size))

                dst_band = dst_ds.GetRasterBand(dst_index)
                if src_band.GetNoDataValue() is not None:
                    dst_band.SetNoDataValue(src_band.GetNoDataValue())

                for block_x, block_y in product(range(num_x), range(num_y)):
                    offset_x = block_x * block_x_size
                    offset_y = block_y * block_y_size
                    size_x = min(src_band.XSize - offset_x, block_x_size)
                    size_y = min(src_band.YSize - offset_y, block_y_size)
                    data = src_band.ReadAsArray(
                        offset_x, offset_y, size_x, size_y
                    )

                    # perform clipping and scaling
                    data = ((dst_range[1] - dst_range[0]) *
                            ((numpy.clip(data, dmin, dmax) - src_range[0]) /
                            (src_range[1] - src_range[0])))

                    # set new datatype
                    data = data.astype(gdal_array.codes[self.datatype])

                    # write result
                    dst_band.WriteArray(data, offset_x, offset_y)

                    # write equal bands at once
                    if multiple > 0:
                        for i in range(multiple):
                            dst_band_multiple = dst_ds.GetRasterBand(
                                dst_index-1-i
                            )
                            dst_band_multiple.WriteArray(
                                data, offset_x, offset_y
                            )
                        multiple_written = True

                if multiple_written:
                    multiple = 0
                    multiple_written = False

            copy_projection(src_ds, dst_ds)
            copy_metadata(src_ds, dst_ds)

            return dst_ds

        except:
            cleanup_temp(dst_ds)
            raise


class ColorIndexOptimization(DatasetOptimization):
    """ Dataset optimization step to replace the pixel color values with a color
        index. If no color palette is given (e.g: a VRT or any other dataset
        containing a color table), this step takes the first three bands and
        computes a median color table.
    """

    def __init__(self, palette_file=None, temporary_directory=None):
        self.palette_file = palette_file
        self.temporary_directory = temporary_directory

    def __call__(self, src_ds):
        logger.info("Applying ColorIndexOptimization")
        try:
            dst_ds = create_temp(src_ds.RasterXSize, src_ds.RasterYSize,
                                 1, gdal.GDT_Byte,
                                 temp_root=self.temporary_directory)

            if not self.palette_file:
                # create a color table as a median of the given dataset
                ct = gdal.ColorTable()
                gdal.ComputeMedianCutPCT(src_ds.GetRasterBand(1),
                                         src_ds.GetRasterBand(2),
                                         src_ds.GetRasterBand(3),
                                         256, ct)

            else:
                # copy the color table from the given palette file
                pct_ds = gdal.Open(self.palette_file)
                pct_ct = pct_ds.GetRasterBand(1).GetRasterColorTable()
                if not pct_ct:
                    raise ValueError("The palette file '%s' does not have a "
                                     "Color Table." % self.palette_file)
                ct = pct_ct.Clone()
                pct_ds = None

            dst_ds.GetRasterBand(1).SetRasterColorTable(ct)
            gdal.DitherRGB2PCT(src_ds.GetRasterBand(1),
                               src_ds.GetRasterBand(2),
                               src_ds.GetRasterBand(3),
                               dst_ds.GetRasterBand(1), ct)

            copy_projection(src_ds, dst_ds)
            copy_metadata(src_ds, dst_ds)

            return dst_ds
        except:
            cleanup_temp(dst_ds)
            raise


class NoDataValueOptimization(DatasetOptimization):
    """ This optimization step assigns a no-data value to all raster bands in
        a dataset.
    """

    def __init__(self, nodata_values):
        self.nodata_values = nodata_values

    def __call__(self, ds):
        logger.info("Applying NoDataValueOptimization")
        nodata_values = self.nodata_values
        if len(nodata_values) == 1:
            nodata_values = nodata_values * ds.RasterCount

        #TODO: bug, the same nodata value is set to all bands?

        for index, value in enumerate(nodata_values, start=1):
            try:
                ds.GetRasterBand(index).SetNoDataValue(value)
            except RuntimeError:
                pass  # TODO

        return ds


#===============================================================================
# Post-create optimization steps
#===============================================================================

class DatasetPostOptimization(object):
    """ Abstract base class for dataset post-creation optimization steps. These
        opotimizations are performed on the actually produced dataset. This is
        required by some optimization techiques.
    """

    def __call__(self, ds):
        raise NotImplementedError


class OverviewOptimization(DatasetPostOptimization):
    """ Dataset optimization step to add overviews to the dataset. This step may
        have to be applied after the dataset has been reprojected.
    """

    def __init__(self, resampling=None, levels=None, minsize=None):
        self.resampling = resampling
        self.levels = levels
        self.minsize = minsize

    def __call__(self, ds):
        logger.info("Applying OverviewOptimization")
        levels = self.levels

        # calculate the overviews automatically.
        if not levels:
            desired_size = abs(self.minsize or 256)
            size = max(ds.RasterXSize, ds.RasterYSize)
            level = 1
            levels = []

            while size > desired_size:
                size /= 2
                level *= 2
                levels.append(level)

        logger.info("Building overview levels %s with resampling method '%s'."
                    % (", ".join(map(str, levels)), self.resampling))

        # drop previous overviews
        ds.BuildOverviews(self.resampling, [])

        # re-build overviews
        # workaround for libtiff 3.X systems, which generated wrong overviews
        # on some levels. Skip with warning if workaround is not working.
        for level in levels:
            try:
                ds.BuildOverviews(self.resampling, [level])
            except RuntimeError:
                logger.warning("Overview building failed for level '%s'." %
                               level)

        return ds


#===============================================================================
# AlphaBand Optimization
#===============================================================================

class AlphaBandOptimization(object):
    """ This optimization renders the footprint into the alpha channel of the
    image. """

    def __call__(self, src_ds, footprint_wkt):
        logger.info("Applying AlphaBandOptimization")
        dt = src_ds.GetRasterBand(1).DataType
        if src_ds.RasterCount == 3:
            src_ds.AddBand(dt)
        elif src_ds.RasterCount == 4:
            pass  # okay
        else:
            raise Exception("Cannot add alpha band, as the current number of "
                            "bands '%d' does not match" % src_ds.RasterCount)

        # initialize the alpha band with zeroes (completely transparent)
        band = src_ds.GetRasterBand(4)
        band.Fill(0)

        # set up the layer with geometry
        ogr_ds = ogr.GetDriverByName('Memory').CreateDataSource('wkt')

        sr = osr.SpatialReference()
        sr.ImportFromEPSG(4326)
        layer = ogr_ds.CreateLayer('poly', srs=sr.sr)

        feat = ogr.Feature(layer.GetLayerDefn())
        feat.SetGeometryDirectly(ogr.Geometry(wkt=footprint_wkt))
        layer.CreateFeature(feat)

        # rasterize the polygon, burning the opaque value into the alpha band
        gdal.RasterizeLayer(src_ds, [4], layer, burn_values=[get_limits(dt)[1]])


#===============================================================================
# Color To Alpha Optimization
#===============================================================================

class ColorToAlphaOptimization(DatasetOptimization):
    """
    Pixel-wise optimization burns color_to_alpha pixels of RGB 
    into alpha channel as 0 (transparent) if all bands have pixel value = [color_to_alpha +- margin].
    Default black is turned to transparent. Returns modified dataset.
    """
    def __call__(self, src_ds, color_to_alpha=0, color_to_alpha_margin=15):
        logger.info("Applying ColorToAlphaOptimization")
        dt = src_ds.GetRasterBand(1).DataType
        higher_cut = color_to_alpha + color_to_alpha_margin
        lower_cut = color_to_alpha - color_to_alpha_margin
        largest_value_of_datatype = get_limits(dt)[1]
        if src_ds.RasterCount == 3:
            src_ds.AddBand(dt)
            # prefill with not-transparent
            src_ds.GetRasterBand(4).fill(largest_value_of_datatype)
        elif src_ds.RasterCount == 4:
            pass  # okay
        else:
            raise Exception("Cannot add alpha band, as the current number of "
                            "bands '%d' does not match" % src_ds.RasterCount)
        try:
            # save all bands to memory -> np array
            as_array_3d = src_ds.ReadAsArray()
            # set alpha pixel to transparent if all three are that color
            [R, G, B, A] = numpy.split(as_array_3d, 4)
            # turn to two dimensional arrays
            R = R[0, :, :]
            G = G[0, :, :]
            B = B[0, :, :]
            A = A[0, :, :]
            # compute masks
            R_mask = numpy.zeros(R.shape, dtype=bool)
            G_mask = numpy.zeros(R.shape, dtype=bool)
            B_mask = numpy.zeros(R.shape, dtype=bool)
            R_mask = (R < higher_cut) & (R > lower_cut)
            G_mask = (G < higher_cut) & (G > lower_cut)
            B_mask = (B < higher_cut) & (B > lower_cut)
            # merge three separate masks with logical AND
            rg_mask = numpy.logical_and(R_mask, G_mask)
            rgb_mask = numpy.logical_and(rg_mask, B_mask)
            # insert into alpha channel
            A[rgb_mask] = 0
            src_ds.GetRasterBand(4).WriteArray(A)
            return src_ds
        except:
            raise



#!/usr/bin/env python
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import sys, os
from lxml import etree

"""
    Validate given XML document(s) against schema(s) referenced in 
    schemaLocation attribute of root element.
    
    Usage: validate_xml.py <xml-filename>
    
    Basically same as: xmllint --noout --schema <schema-filename> <xml-filename>
"""
if __name__ == "__main__" :
    # Validate all given files
    for name in sys.argv:
        # but not script itself
        if name == sys.argv[0]:
            continue
        # Check if file exists
        if not os.path.exists(name):
            print ("Usage: %s <xml-filename>" % sys.argv[0])
            sys.exit(1)

        print ("Validating XML document: %s" % name)
        doc = etree.parse(name).getroot()
        schema_locations = doc.get("{http://www.w3.org/2001/XMLSchema-instance}schemaLocation")
        locations = schema_locations.split()
        
        # Create schema importing all referenced schemas
        schema_def = etree.Element("schema", attrib={
                "elementFormDefault": "qualified",
                "version": "1.0.0",
            }, nsmap={
                None: "http://www.w3.org/2001/XMLSchema"
            }
        )
        for ns, location in zip(locations[::2], locations[1::2]):
            etree.SubElement(schema_def, "import", attrib={
                    "namespace": ns,
                    "schemaLocation": location
                }
            )
            print ("Schema: Namespace: %s, Location: %s" % (ns, location))
        
        # TODO: Workaround, but otherwise the doc is not recognized as schema
        schema = etree.XMLSchema(etree.XML(etree.tostring(schema_def)))
        
        # Validate file against schema
        try:
            schema.assertValid(doc)
            print ("%s validates" % name)
        except etree.Error as e:
            print ("%s doesn't validate" % name)
            print ("Error was: %s" % str(e))



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage browse types. This command uses sub-commands for the
        specific tasks: create, delete, list
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        list_parser = self.add_subparser(parser, 'list')

        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'product_type_name', nargs=1,
                help='The product type name. Mandatory.'
            )

            parser.add_argument(
                'browse_type_name', nargs='?', default='',
                help='The browse type name. Optional.'
            )

        create_parser.add_argument(
            '--red', '-r', '--grey',
            dest='red_or_grey_expression', default=None,
        )
        create_parser.add_argument(
            '--green', '-g',
            dest='green_expression', default=None,
        )
        create_parser.add_argument(
            '--blue', '-b',
            dest='blue_expression', default=None,
        )
        create_parser.add_argument(
            '--alpha', '-a',
            dest='alpha_expression', default=None,
        )

        create_parser.add_argument(
            '--red-range', '--grey-range', nargs=2, type=float,
            dest='red_or_grey_range', default=(None, None),
        )
        create_parser.add_argument(
            '--green-range', nargs=2, type=float,
            dest='green_range', default=(None, None),
        )
        create_parser.add_argument(
            '--blue-range', nargs=2, type=float,
            dest='blue_range', default=(None, None),
        )
        create_parser.add_argument(
            '--alpha-range', nargs=2, type=float,
            dest='alpha_range', default=(None, None),
        )

        create_parser.add_argument(
            '--red-nodata', '--grey-nodata', type=float,
            dest='red_or_grey_nodata', default=None,
        )
        create_parser.add_argument(
            '--green-nodata', type=float,
            dest='green_nodata', default=None,
        )
        create_parser.add_argument(
            '--blue-nodata', type=float,
            dest='blue_nodata', default=None,
        )
        create_parser.add_argument(
            '--alpha-nodata', type=float,
            dest='alpha_nodata', default=None,
        )

        list_parser.add_argument(
            'product_type_name', nargs=1,
            help='The product type name. Mandatory.'
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: create, delete.
        """
        if subcommand == "create":
            self.handle_create(
                kwargs.pop('product_type_name')[0], *args, **kwargs
            )
        elif subcommand == "delete":
            self.handle_delete(
                kwargs.pop('product_type_name')[0], *args, **kwargs
            )
        elif subcommand == "list":
            self.handle_list(
                kwargs.pop('product_type_name')[0], *args, **kwargs
            )

    def handle_create(self, product_type_name, browse_type_name,
                      red_or_grey_expression, green_expression,
                      blue_expression, alpha_expression,
                      red_or_grey_range=(None, None), green_range=(None, None),
                      blue_range=(None, None), alpha_range=(None, None),
                      red_or_grey_nodata=None, green_nodata=None,
                      blue_nodata=None, alpha_nodata=None,
                      *args, **kwargs):
        """ Handle the creation of a new browse type.
        """

        try:
            product_type = models.ProductType.objects.get(
                name=product_type_name
            )
        except models.ProductType.DoesNotExist:
            raise CommandError(
                'Product type %r does not exist' % product_type_name
            )

        red_min, red_max = red_or_grey_range
        green_min, green_max = green_range
        blue_min, blue_max = blue_range
        alpha_min, alpha_max = alpha_range

        models.BrowseType.objects.create(
            product_type=product_type,
            name=browse_type_name,
            red_or_grey_expression=red_or_grey_expression,
            green_expression=green_expression,
            blue_expression=blue_expression,
            alpha_expression=alpha_expression,
            red_or_grey_range_min=red_min,
            red_or_grey_range_max=red_max,
            green_range_min=green_min,
            green_range_max=green_max,
            blue_range_min=blue_min,
            blue_range_max=blue_max,
            alpha_range_min=alpha_min,
            alpha_range_max=alpha_max,
            red_or_grey_nodata_value=red_or_grey_nodata,
            green_nodata_value=green_nodata,
            blue_nodata_value=blue_nodata,
            alpha_nodata_value=alpha_nodata,
        )

        if not browse_type_name:
            print(
                'Successfully created default browse type for product_type %r'
                % product_type_name
            )
        else:
            print(
                'Successfully created browse type %r for product_type %r'
                % (browse_type_name, product_type_name)
            )

    def handle_delete(self, product_type_name, browse_type_name, **kwargs):
        """ Handle the deletion of a browse type
        """

        try:
            product_type = models.ProductType.objects.get(
                name=product_type_name
            )
        except models.ProductType.DoesNotExist:
            raise CommandError('No such product type %r' % product_type_name)

        browse_type = product_type.browse_types.get(name=browse_type_name)

        browse_type.delete()

        if not browse_type_name:
            print(
                'Successfully deleted default browse type for product_type %r'
                % product_type_name
            )
        else:
            print(
                'Successfully deleted browse type %r for product_type %r'
                % (browse_type_name, product_type_name)
            )

    def handle_list(self, product_type_name, *args, **kwargs):
        """ Handle the listing of browse types
        """
        try:
            product_type = models.ProductType.objects.get(
                name=product_type_name
            )
        except models.ProductType.DoesNotExist:
            raise CommandError('No such product type %r' % product_type_name)

        for browse_type in product_type.browse_types.all():
            print(browse_type.name or '(Default)')

            red = browse_type.red_or_grey_expression
            green = browse_type.green_expression
            blue = browse_type.blue_expression
            alpha = browse_type.alpha_expression

            if red and not green and not blue and not alpha:
                print('\tGrey: \'%s\'' % red)

            if red:
                print('\tRed: \'%s\'' % red)

            if green:
                print('\tGreen: \'%s\'' % green)

            if blue:
                print('\tBlue: \'%s\'' % blue)

            if alpha:
                print('\tAlpha: \'%s\'' % alpha)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.contrib.gis.geos import Polygon
from django.contrib.gis.gdal import SpatialReference, CoordTransform, DataSource

from eoxserver.contrib import gdal
from eoxserver.backends.access import get_vsi_path, get_vsi_env, gdal_open
from eoxserver.render.coverage.objects import Coverage


BROWSE_MODE_RGB = "rgb"
BROWSE_MODE_RGBA = "rgba"
BROWSE_MODE_GRAYSCALE = "grayscale"


class Browse(object):
    def __init__(self, name, filename, env, size, extent, crs, mode, footprint):
        self._name = name
        self._filename = filename
        self._env = env
        self._size = size
        self._extent = extent
        self._crs = crs
        self._mode = mode
        self._footprint = footprint

    @property
    def name(self):
        return self._name

    @property
    def filename(self):
        return self._filename

    @property
    def env(self):
        return self._env

    @property
    def size(self):
        return self._size

    @property
    def extent(self):
        return self._extent

    @property
    def crs(self):
        return self._crs

    @property
    def spatial_reference(self):
        return SpatialReference(self.crs)

    @property
    def mode(self):
        return self._mode

    @property
    def footprint(self):
        if self._footprint:
            return self._footprint
        else:
            polygon = Polygon.from_bbox(self.extent)
            srs = SpatialReference(self.crs)
            if srs.srid != 4326:
                ct = CoordTransform(srs, SpatialReference(4326))
                polygon.transform(ct)
            return polygon

    @classmethod
    def from_model(cls, product_model, browse_model):
        filename = get_vsi_path(browse_model)
        env = get_vsi_env(browse_model.storage)
        size = (browse_model.width, browse_model.height)
        extent = (
            browse_model.min_x, browse_model.min_y,
            browse_model.max_x, browse_model.max_y
        )

        ds = gdal_open(browse_model)
        mode = _get_ds_mode(ds)
        ds = None

        if browse_model.browse_type:
            name = '%s__%s' % (
                product_model.identifier, browse_model.browse_type.name
            )
        else:
            name = product_model.identifier

        return cls(
            name, filename, env, size, extent,
            browse_model.coordinate_reference_system, mode,
            product_model.footprint
        )

    @classmethod
    def from_file(cls, filename, env=None):
        env = env or {}
        ds = gdal.Open(filename)
        size = (ds.RasterXSize, ds.RasterYSize)
        extent = gdal.get_extent(ds)
        mode = _get_ds_mode(ds)

        return cls(
            filename, env, filename, size, extent,
            ds.GetProjection(), mode, None
        )


class GeneratedBrowse(Browse):
    def __init__(self, name, band_expressions, ranges, fields_and_coverages,
                 field_list, footprint):
        self._name = name
        self._band_expressions = band_expressions
        self._ranges = ranges
        self._fields_and_coverages = fields_and_coverages
        self._field_list = field_list
        self._footprint = footprint

    @property
    def name(self):
        return self._name

    @property
    def size(self):
        for field, coverages in self._fields_and_coverages.items():
            return coverages[0].size

    @property
    def extent(self):
        for field, coverages in self._fields_and_coverages.items():
            return coverages[0].extent

    @property
    def crs(self):
        for field, coverages in self._fields_and_coverages.items():
            return coverages[0].grid.coordinate_reference_system

    @property
    def spatial_reference(self):
        for field, coverages in self._fields_and_coverages.items():
            return coverages[0].grid.spatial_reference

    @property
    def mode(self):
        field_count = len(self._band_expressions)
        if field_count == 1:
            return BROWSE_MODE_GRAYSCALE
        elif field_count == 3:
            return BROWSE_MODE_RGB
        elif field_count == 4:
            return BROWSE_MODE_RGB

    @property
    def band_expressions(self):
        return self._band_expressions

    @property
    def ranges(self):
        return self._ranges

    @property
    def fields_and_coverages(self):
        return self._fields_and_coverages

    @property
    def field_list(self):
        return self._field_list

    @classmethod
    def from_coverage_models(cls, band_expressions_and_ranges,
                             fields_and_coverage_models,
                             product_model):

        band_expressions, ranges = zip(*band_expressions_and_ranges)

        fields_and_coverages = {
            field_name: [
                Coverage.from_model(coverage)
                for coverage in coverages
            ]
            for field_name, coverages in fields_and_coverage_models.items()
        }

        return cls(
            product_model.identifier,
            band_expressions,
            ranges,
            fields_and_coverages, [
                fields_and_coverages[field_name][0].range_type.get_field(
                    field_name
                )
                for field_name in fields_and_coverages.keys()
            ],
            product_model.footprint
        )


class Mask(object):
    def __init__(self, filename=None, geometry=None, validity=False):
        self._filename = filename
        self._geometry = geometry
        self._validity = validity

    @property
    def filename(self):
        return self._filename

    @property
    def geometry(self):
        return self._geometry

    def load_geometry(self):
        ds = DataSource(self.filename)
        layer = ds[0]
        geometries = layer.get_geoms()

        first = geometries[0]
        for other in geometries[1:]:
            first = first.union(other)
        return first.geos

    @property
    def validity(self):
        return self._validity

    @classmethod
    def from_model(cls, mask_model, mask_type):
        filename = None
        if mask_model and mask_model.location:
            filename = get_vsi_path(mask_model)

        geometry = None
        if mask_model:
            geometry = mask_model.geometry

        mask_type = mask_type or mask_model.mask_type

        validity = False
        if mask_type:
            validity = mask_type.validity

        return cls(filename, geometry, validity)


class MaskedBrowse(object):
    def __init__(self, browse, mask):
        self._browse = browse
        self._mask = mask

    @property
    def browse(self):
        return self._browse

    @property
    def mask(self):
        return self._mask

    @classmethod
    def from_models(cls, product_model, browse_model, mask_model,
                    mask_type_model):
        return cls(
            Browse.from_model(product_model, browse_model),
            Mask.from_model(mask_model, mask_type_model)
        )


def _get_ds_mode(ds):
    first = ds.GetRasterBand(1)

    count = ds.RasterCount
    if count == 1 or count > 4 and not first.GetColorTable():
        mode = BROWSE_MODE_GRAYSCALE
    elif (count == 1 and first.GetColorTable()) or count == 4:
        mode = BROWSE_MODE_RGBA
    elif count == 3 and first.GetColorInterpretation() == gdal.GCI_RedBand:
        mode = BROWSE_MODE_RGB

    return mode



import logging
import os
import datetime
import httplib
import xml.dom.minidom
import eoxserver
from urlparse import urlparse

from django.utils.six import iteritems

from eoxserver.core import implements
from eoxserver.services.auth.base import BasePDP
from eoxserver.services.auth.interfaces import PolicyDecisionPointInterface
from eoxserver.services.auth.charonpdp import CharonPDP


logger = logging.getLogger(__name__)

validUser = {
    'uid': 'jdoe', 
    'cn': 'Doe John', 
    'sn': 'Doe', 
    'description': 'Authorized User'
}

class DummyPDP(BasePDP):
    implements(PolicyDecisionPointInterface)

    def __init__(self):
        self.pdp = CharonPDP(DummyAuthzClient())

    def _decide(self, request):
        httpHeader = request.META

        #checks if a attribute 'DUMMY_MODE' is in the headers
        if 'DUMMY_MODE' in httpHeader:
            logger.info("Security Test: 'DUMMY_MODE' parameter in HTTP header")
            return self.pdp._decide(request)
        else :
            return (True, 'No authorisation testing')


class DummyAuthzClient(object):
    def authorize(self, userAttributes, resourceAttributes, action):
        for key, value in iteritems(validUser):
            if key in userAttributes:
                if value != userAttributes[key]:
                    return (False, 'Not Authorised')
            else:
                return (False, 'Not Authorised')

        return (True, 'Authorised')



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011-2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import os.path

from django.conf import settings

from eoxserver.core import Component, implements
from eoxserver.contrib import mapserver as ms
from eoxserver.contrib import gdal
from eoxserver.resources.coverages import models, crss
from eoxserver.services.mapserver.interfaces import LayerFactoryInterface
from eoxserver.services import models as service_models

from eoxserver.resources.coverages.dateline import (
    extent_crosses_dateline, wrap_extent_around_dateline
)

#-------------------------------------------------------------------------------


class BaseStyleMixIn(object):
    STYLES = (
        ("red", 255, 0, 0),
        ("green", 0, 128, 0),
        ("blue", 0, 0, 255),
        ("white", 255, 255, 255),
        ("black", 0, 0, 0),
        ("yellow", 255, 255, 0),
        ("orange", 255, 165, 0),
        ("magenta", 255, 0, 255),
        ("cyan", 0, 255, 255),
        ("brown", 165, 42, 42)
    )

    DEFAULT_STYLE = "red"

    def apply_styles(self, layer, fill=False):
        # add style info
        for name, r, g, b in self.STYLES:
            cls = ms.classObj()
            style = ms.styleObj()
            style.outlinecolor = ms.colorObj(r, g, b)
            if fill:
                style.color = ms.colorObj(r, g, b)
            cls.insertStyle(style)
            cls.group = name

            layer.insertClass(cls)

        layer.classgroup = self.DEFAULT_STYLE


class OffsiteColorMixIn(object):
    def offsite_color_from_range_type(self, range_type, band_indices=None):
        """ Helper function to create an offsite color for a given range type
            and optionally band indices.
        """
        if band_indices is None:
            if len(range_type) == 1:
                band_indices = [0, 0, 0]
            elif len(range_type) >= 3:
                band_indices = [0, 1, 2]
            else:
                # no offsite color possible
                return None

        if len(band_indices) != 3:
            raise ValueError(
                "Wrong number of band indices to calculate offsite color."
            )

        values = []
        for index in band_indices:
            band = range_type[index]
            nil_value_set = band.nil_value_set

            # we only support offsite colors for "Byte" bands
            if nil_value_set and nil_value_set.data_type != gdal.GDT_Byte:
                return None

            if nil_value_set and len(nil_value_set) > 0:
                values.append(nil_value_set[0].value)
            else:
                return None

        return ms.colorObj(*values)


class PolygonLayerMixIn(object):
    def _create_polygon_layer(self, name):
        layer = ms.layerObj()
        layer.name = name
        layer.type = ms.MS_LAYER_POLYGON

        self.apply_styles(layer)

        srid = 4326
        layer.setProjection(crss.asProj4Str(srid))
        layer.setMetaData("ows_srs", crss.asShortCode(srid))
        layer.setMetaData("wms_srs", crss.asShortCode(srid))

        layer.dump = True

        layer.header = os.path.join(settings.PROJECT_DIR, "conf", "outline_template_header.html")
        layer.template = os.path.join(settings.PROJECT_DIR, "conf", "outline_template_dataset.html")
        layer.footer = os.path.join(settings.PROJECT_DIR, "conf", "outline_template_footer.html")

        layer.setMetaData("gml_include_items", "all")
        layer.setMetaData("wms_include_items", "all")

        layer.addProcessing("ITEMS=identifier")

        layer.offsite = ms.colorObj(0, 0, 0)

        return layer


class PlainLayerMixIn(object):

    def _create_layer(self, coverage, name, extent=None, group=None,
                      wrapped=False):
        layer = ms.layerObj()
        layer.name = name
        layer.type = ms.MS_LAYER_RASTER
        if extent:
            layer.setMetaData("wms_extent", "%f %f %f %f" % extent)
            layer.setExtent(*extent)

        #layer.setMetaData(
        #    "wms_enable_request", "getcapabilities getmap getfeatureinfo"
        #)

        if wrapped:
            # set the info for the connector to wrap this layer around the
            # dateline
            layer.setMetaData("eoxs_wrap_dateline", "true")

        self._set_projection(layer, coverage.spatial_reference)
        if group:
            layer.group = group

        return layer

    def get_render_options(self, coverage):
        try:
            return service_models.WMSRenderOptions.objects.get(
                coverage=coverage
            )
        except service_models.WMSRenderOptions.DoesNotExist:
            return None

    def set_render_options(self, layer, offsite=None, options=None):
        if offsite:
            layer.offsite = offsite
        if options:
            red = options.default_red
            green = options.default_green
            blue = options.default_blue
            alpha = options.default_alpha

            # rgb(a) bands
            if red is not None and green is not None and blue is not None:
                bands = "%d,%d,%d" % (red, green, blue)
                if alpha is not None:
                    bands += alpha
                layer.setProcessingKey("BANDS", bands)

            # single band grey
            elif red is not None and (green, blue, alpha) == (None, None, None):
                layer.setProcessingKey("BANDS", str(red))

            if options.bands_scale_min and options.bands_scale_max:
                bands_scale_min = str(options.bands_scale_min).split(',')
                bands_scale_max = str(options.bands_scale_max).split(',')

                if red is not None and (green, blue) == (None, None):
                    idx1 = red - 1
                    layer.setProcessingKey("SCALE", "%d,%d" % (
                        bands_scale_min[idx1], bands_scale_max[idx1]
                    ))
                else:
                    idx1 = (red or 1) - 1
                    idx2 = (green or 2) - 1
                    idx3 = (blue or 3) - 1
                    layer.setProcessingKey("SCALE_1", "%s,%s" % (
                        bands_scale_min[idx1], bands_scale_max[idx1]
                    ))
                    layer.setProcessingKey("SCALE_2", "%s,%s" % (
                        bands_scale_min[idx2], bands_scale_max[idx2]
                    ))
                    layer.setProcessingKey("SCALE_3", "%s,%s" % (
                        bands_scale_min[idx3], bands_scale_max[idx3]
                    ))

            elif options.scale_auto:
                layer.setProcessingKey("SCALE", "AUTO")
            elif options.scale_min is not None and options.scale_max is not None:
                layer.setProcessingKey(
                    "SCALE", "%d,%d" % (options.scale_min, options.scale_max)
                )

            if options.resampling:
                layer.setProcessingKey("RESAMPLE", str(options.resampling))

    def _set_projection(self, layer, sr):
        if sr.srid is not None:
            short_epsg = "EPSG:%d" % sr.srid
            layer.setMetaData("ows_srs", short_epsg)
            layer.setMetaData("wms_srs", short_epsg)
        layer.setProjection(sr.proj)


class AbstractLayerFactory(PlainLayerMixIn, Component):
    implements(LayerFactoryInterface)
    abstract = True

    def generate(self, eo_object, group_layer, suffix, options):
        raise NotImplementedError


class BaseCoverageLayerFactory(OffsiteColorMixIn, PlainLayerMixIn, Component):
    implements(LayerFactoryInterface)
    # NOTE: Technically, this class is not abstract but we need it labeled
    #       as an abstract class to allow inheritance while avoiding multiple
    #       component's imports.
    abstract = True

    def generate(self, eo_object, group_layer, suffix, options):
        coverage = eo_object.cast()
        extent = coverage.extent
        srid = coverage.srid

        data_items = coverage.data_items.all()
        range_type = coverage.range_type

        offsite = self.offsite_color_from_range_type(range_type)
        options = self.get_render_options(coverage)

        if srid is not None and extent_crosses_dateline(extent, srid):
            identifier = coverage.identifier
            wrapped_extent = wrap_extent_around_dateline(extent, srid)
            layer = self._create_layer(
                coverage, identifier + "_unwrapped", extent, identifier
            )

            self.set_render_options(layer, offsite, options)
            self.set_render_options(layer, offsite, options)
            yield layer, data_items
            wrapped_layer = self._create_layer(
                coverage, identifier + "_wrapped", wrapped_extent, identifier, True
            )
            self.set_render_options(wrapped_layer, offsite, options)
            yield wrapped_layer, data_items
        else:
            layer = self._create_layer(
                coverage, coverage.identifier, extent
            )
            self.set_render_options(layer, offsite, options)
            yield layer, data_items

    def generate_group(self, name):
        return ms.Layer(name)
